import functools
import sys
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from types import TracebackType
from typing import TYPE_CHECKING, Any, NamedTuple, TypeAlias, TypeVar, Union, cast, overload

import dagster._check as check
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.assets.graph.base_asset_graph import EntityKey
from dagster._core.definitions.assets.graph.remote_asset_graph import (
    RemoteAssetCheckNode,
    RemoteAssetNode,
    RemoteWorkspaceAssetGraph,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.selector import (
    GraphSelector,
    JobSelector,
    JobSubsetSelector,
    RepositorySelector,
    ScheduleSelector,
    SensorSelector,
)
from dagster._core.definitions.temporal_context import TemporalContext
from dagster._core.errors import DagsterError, DagsterInvariantViolationError
from dagster._core.execution.backfill import PartitionBackfill
from dagster._core.remote_representation.code_location import is_implicit_asset_job_name
from dagster._core.remote_representation.external import RemoteJob, RemoteSchedule, RemoteSensor
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.workspace.context import RemoteDefinition
from dagster._core.workspace.permissions import Permissions
from dagster._utils.caching_instance_queryer import CachingInstanceQueryer
from dagster._utils.error import serializable_error_info_from_exc_info
from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from dagster._core.workspace.context import BaseWorkspaceRequestContext

    from dagster_graphql.schema.errors import GrapheneError, GraphenePythonError
    from dagster_graphql.schema.util import ResolveInfo

P = ParamSpec("P")
T = TypeVar("T")

GrapheneResolverFn: TypeAlias = Callable[..., object]
T_Callable = TypeVar("T_Callable", bound=Callable)


def assert_permission_for_location(
    graphene_info: "ResolveInfo", permission: str, location_name: str
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    context = cast("BaseWorkspaceRequestContext", graphene_info.context)
    if not context.has_permission_for_location(permission, location_name):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def require_permission_check(
    permission: str,
):
    def decorator(fn):
        if iscoroutinefunction(fn):

            @functools.wraps(fn)
            async def _async_fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
                result = await fn(self, graphene_info, *args, **kwargs)
                if not graphene_info.context.was_permission_checked(permission):
                    raise Exception(f"Permission {permission} was never checked during the request")
                return result

            return _async_fn
        else:

            @functools.wraps(fn)
            def _fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
                result = fn(self, graphene_info, *args, **kwargs)

                if not graphene_info.context.was_permission_checked(permission):
                    raise Exception(f"Permission {permission} was never checked during the request")

                return result

            return _fn

    return decorator


def check_permission(
    permission: str,
):
    def decorator(fn):
        if iscoroutinefunction(fn):

            @functools.wraps(fn)
            async def _async_fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
                assert_permission(graphene_info, permission)

                return await fn(self, graphene_info, *args, **kwargs)

            return _async_fn
        else:

            @functools.wraps(fn)
            def _fn(self, graphene_info, *args: P.args, **kwargs: P.kwargs):
                assert_permission(graphene_info, permission)

                return fn(self, graphene_info, *args, **kwargs)

            return _fn

    return decorator


def assert_permission(graphene_info: "ResolveInfo", permission: str) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    context = cast("BaseWorkspaceRequestContext", graphene_info.context)
    if not context.has_permission(permission):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def has_permission_for_asset_graph(
    graphene_info: "ResolveInfo",
    asset_graph: RemoteWorkspaceAssetGraph,
    entity_keys: Sequence[EntityKey] | None,
    permission: str,
) -> bool:
    all_keys = set(entity_keys) if entity_keys else set()
    context = cast("BaseWorkspaceRequestContext", graphene_info.context)

    # if we have the permission for the whole deployment, no need to check specific asset keys or locations
    if context.has_permission(permission):
        return True

    if (
        not any(
            context.has_permission_for_location(permission, location_name)
            for location_name in context.code_location_names
        )
        and not context.viewer_has_any_owner_definition_permissions()
    ):
        # short-circuit if we don't have any location-level permissions or definition-level permissions
        return False

    location_names = set()
    if all_keys:
        for key in all_keys:
            if not asset_graph.has(key):
                # If any of the asset keys don't map to a location (e.g. because they are no longer in the
                # graph) need deployment-wide permissions - no valid code location to check
                return context.has_permission(permission)
            location_name = asset_graph.get_repository_handle(key).location_name
            location_names.add(location_name)
    else:
        location_names = set(
            handle.location_name for handle in asset_graph.repository_handles_by_key.values()
        )

    if not location_names:
        return context.has_permission(permission)

    # if we have permission for all locations relevant to the asset graph, we're good
    if all(
        context.has_permission_for_location(permission, location_name)
        for location_name in location_names
    ):
        return True

    # No need to check individual asset keys if we don't have owner permissions
    if not context.viewer_has_any_owner_definition_permissions():
        return False

    if not all_keys:
        return False

    return all(
        context.has_permission_for_selector(permission, entity_key) for entity_key in all_keys
    )


def assert_permission_for_asset_graph(
    graphene_info: "ResolveInfo",
    asset_graph: RemoteWorkspaceAssetGraph,
    entity_keys: Sequence[EntityKey] | None,
    permission: str,
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not has_permission_for_asset_graph(graphene_info, asset_graph, entity_keys, permission):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def has_permission_for_definition(
    graphene_info: "ResolveInfo", permission: str, remote_definition: RemoteDefinition
):
    if graphene_info.context.has_permission(permission):
        return True

    location_name = location_name_for_remote_definition(remote_definition)
    if graphene_info.context.has_permission_for_location(permission, location_name):
        return True

    if not graphene_info.context.viewer_has_any_owner_definition_permissions():
        return False

    if (
        isinstance(remote_definition, RemoteJob)
        and is_implicit_asset_job_name(remote_definition.name)
        and remote_definition.asset_selection
    ):
        return has_permission_for_asset_graph(
            graphene_info,
            graphene_info.context.asset_graph,
            list(remote_definition.asset_selection),
            permission,
        )

    if isinstance(remote_definition, RemoteAssetCheckNode):
        owners = graphene_info.context.get_owners_for_selector(remote_definition.asset_check.key)
    else:
        owners = remote_definition.owners

    if not owners:
        return False

    return graphene_info.context.has_permission_for_owners(permission, owners)


def location_name_for_remote_definition(remote_definition: RemoteDefinition) -> str:
    if isinstance(remote_definition, RemoteAssetNode):
        return (
            remote_definition.resolve_to_singular_repo_scoped_node().repository_handle.location_name
        )
    elif isinstance(
        remote_definition,
        (RemoteJob, RemoteSchedule, RemoteSensor, RemoteAssetCheckNode),
    ):
        return remote_definition.handle.location_name
    else:
        check.failed(f"Unexpected remote definition type {type(remote_definition)}")


def has_permission_for_run(
    graphene_info: "ResolveInfo", permission: Permissions, run: DagsterRun
) -> bool:
    if not run.remote_job_origin:
        return graphene_info.context.has_permission(permission)

    return has_permission_for_job(
        graphene_info,
        permission,
        JobSelector(
            location_name=run.remote_job_origin.location_name,
            repository_name=run.remote_job_origin.repository_origin.repository_name,
            job_name=run.job_name,
        ),
        entity_keys=list(run.entity_selection) if run.entity_selection else None,
    )


def assert_permission_for_run(
    graphene_info: "ResolveInfo", permission: Permissions, run: DagsterRun
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not has_permission_for_run(graphene_info, permission, run):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def has_permission_for_job(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    job_selector: JobSelector,
    entity_keys: Sequence[EntityKey]
    | None = None,  # entity keys are only required for implicit asset jobs
) -> bool:
    if is_implicit_asset_job_name(job_selector.job_name):
        return has_permission_for_asset_graph(
            graphene_info, graphene_info.context.asset_graph, entity_keys, permission
        )

    return graphene_info.context.has_permission_for_selector(permission, job_selector)


def assert_permission_for_job(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    job_selector: JobSelector,
    entity_keys: Sequence[EntityKey]
    | None = None,  # entity keys are only required for implicit asset jobs
):
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not has_permission_for_job(graphene_info, permission, job_selector, entity_keys):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def assert_permission_for_sensor(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    sensor_selector: SensorSelector,
):
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not graphene_info.context.has_permission_for_selector(permission, sensor_selector):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def assert_permission_for_schedule(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    schedule_selector: ScheduleSelector,
):
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not graphene_info.context.has_permission_for_selector(permission, schedule_selector):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def assert_valid_job_partition_backfill(
    graphene_info: "ResolveInfo",
    backfill: PartitionBackfill,
    partitions_def: PartitionsDefinition,
    dynamic_partitions_store: CachingInstanceQueryer,
    backfill_datetime: datetime,
) -> None:
    from dagster_graphql.schema.errors import GraphenePartitionKeysNotFoundError

    partition_names = backfill.get_partition_names(graphene_info.context)

    if not partition_names:
        return

    invalid_keys = set(partition_names) - set(
        partitions_def.get_partition_keys(backfill_datetime, dynamic_partitions_store)
    )

    if invalid_keys:
        raise UserFacingGraphQLError(GraphenePartitionKeysNotFoundError(invalid_keys))


def has_permission_for_backfill(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    backfill: PartitionBackfill,
) -> bool:
    if backfill.is_asset_backfill:
        check.invariant(
            backfill.asset_selection is not None, "Asset backfill must have asset selection"
        )
        return has_permission_for_asset_graph(
            graphene_info,
            graphene_info.context.asset_graph,
            cast("list[AssetKey]", backfill.asset_selection),
            permission,
        )

    # job backfill, check permissions for the job
    if not backfill.partition_set_origin:
        return graphene_info.context.has_permission(permission)

    partition_selector = backfill.partition_set_origin.selector
    if not graphene_info.context.has_code_location_name(partition_selector.location_name):
        return graphene_info.context.has_permission(permission)

    partition_sets = graphene_info.context.get_partition_sets(
        repository_selector=RepositorySelector(
            location_name=partition_selector.location_name,
            repository_name=partition_selector.repository_name,
        )
    )
    matches = [
        partition_set
        for partition_set in partition_sets
        if partition_set.name == partition_selector.partition_set_name
    ]

    if len(matches) != 1:
        return graphene_info.context.has_permission_for_location(
            permission, partition_selector.location_name
        )

    remote_partition_set = next(iter(matches))
    return graphene_info.context.has_permission_for_selector(
        permission,
        JobSelector(
            location_name=partition_selector.location_name,
            repository_name=partition_selector.repository_name,
            job_name=remote_partition_set.job_name,
        ),
    )


def assert_permission_for_backfill(
    graphene_info: "ResolveInfo",
    permission: Permissions,
    backfill: PartitionBackfill,
) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    if not has_permission_for_backfill(graphene_info, permission, backfill):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


def assert_valid_asset_partition_backfill(
    graphene_info: "ResolveInfo",
    backfill: PartitionBackfill,
    backfill_datetime: datetime,
) -> None:
    from dagster_graphql.schema.errors import GraphenePartitionKeysNotFoundError

    asset_graph = graphene_info.context.asset_graph
    asset_backfill_data = backfill.asset_backfill_data

    asset_graph_view = AssetGraphView(
        temporal_context=TemporalContext(
            effective_dt=backfill_datetime,
            last_event_id=None,
        ),
        instance=graphene_info.context.instance,
        asset_graph=asset_graph,
    )

    if not asset_backfill_data:
        return

    checked_execution_set_asset_keys = set()

    for asset_key in asset_backfill_data.target_subset.asset_keys:
        asset_node = asset_graph.get(asset_key)

        entity_subset = asset_graph_view.get_entity_subset_from_asset_graph_subset(
            asset_backfill_data.target_subset, asset_key
        )

        if asset_key not in checked_execution_set_asset_keys:
            checked_execution_set_asset_keys.add(asset_key)
            for execution_set_key in asset_node.execution_set_asset_keys:
                if execution_set_key == asset_key:
                    continue

                if execution_set_key not in asset_backfill_data.target_subset.asset_keys:
                    raise DagsterInvariantViolationError(
                        f"Backfill must include every key in a non-subsettable multi-asset, included "
                        f"{asset_key.to_user_string()} but not {execution_set_key.to_user_string()}"
                    )

                other_entity_subset = asset_graph_view.get_entity_subset_from_asset_graph_subset(
                    asset_backfill_data.target_subset, execution_set_key
                )

                if entity_subset.get_internal_value() != other_entity_subset.get_internal_value():
                    raise DagsterInvariantViolationError(
                        f"Targeted subset must match between every key in"
                        f" a non-subsettable multi-asset, but does not match between {asset_key.to_user_string()} "
                        f"and {execution_set_key.to_user_string()}"
                    )

                checked_execution_set_asset_keys.add(execution_set_key)

        partitions_def = asset_node.partitions_def

        if not partitions_def:
            continue

        invalid_subset = asset_graph_view.get_subset_not_in_graph(
            key=asset_key, candidate_subset=entity_subset
        )

        if not invalid_subset.is_empty:
            raise UserFacingGraphQLError(
                GraphenePartitionKeysNotFoundError(
                    set(invalid_subset.expensively_compute_partition_keys())
                )
            )

        for parent_key in asset_graph.get(asset_key).parent_keys:
            _parent_subset, required_but_nonexistent_subset = (
                asset_graph_view.compute_parent_subset_and_required_but_nonexistent_subset(
                    parent_key,
                    entity_subset,
                )
            )

            if not required_but_nonexistent_subset.is_empty:
                raise DagsterInvariantViolationError(
                    f"Targeted partition subset {entity_subset}"
                    f" depends on non-existent partitions: {required_but_nonexistent_subset}"
                )


def _noop(_) -> None:
    pass


class ErrorCapture:
    @staticmethod
    def default_on_exception(
        exc_info: tuple[type[BaseException], BaseException, TracebackType],
    ) -> "GraphenePythonError":
        from dagster_graphql.schema.errors import GraphenePythonError

        # Transform exception in to PythonError to present to user
        return GraphenePythonError(serializable_error_info_from_exc_info(exc_info))

    # global behavior for how to handle unexpected exceptions
    on_exception = default_on_exception

    # context var for observing unexpected exceptions
    observer: ContextVar[Callable[[Exception], None]] = ContextVar(
        "error_capture_observer", default=_noop
    )

    @staticmethod
    @contextmanager
    def watch(fn: Callable[[Exception], None]) -> Iterator[None]:
        token = ErrorCapture.observer.set(fn)
        try:
            yield
        finally:
            ErrorCapture.observer.reset(token)


@overload
def capture_error(
    fn: Callable[P, T],
) -> Callable[P, T]: ...


@overload
def capture_error(  # pyright: ignore[reportOverlappingOverload]
    fn: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]: ...


def capture_error(
    fn: Callable[P, T] | Callable[P, Awaitable[T]],
) -> (
    Callable[P, Union[T, "GrapheneError", "GraphenePythonError"]]
    | Callable[P, Awaitable[Union[T, "GrapheneError", "GraphenePythonError"]]]
):
    if iscoroutinefunction(fn):

        @functools.wraps(fn)
        async def _async_fn(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return await fn(*args, **kwargs)
            except UserFacingGraphQLError as de_exception:
                return de_exception.error
            except Exception as exc:
                ErrorCapture.observer.get()(exc)
                return ErrorCapture.on_exception(sys.exc_info())  # type: ignore

        return _async_fn

    else:

        @functools.wraps(fn)
        def _fn(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                return cast("T", fn(*args, **kwargs))
            except UserFacingGraphQLError as de_exception:
                return de_exception.error
            except Exception as exc:
                ErrorCapture.observer.get()(exc)
                return ErrorCapture.on_exception(sys.exc_info())  # type: ignore

        return _fn


class UserFacingGraphQLError(Exception):
    # The `error` arg here should be a Graphene type implementing the interface `GrapheneError`, but
    # this is not trackable by the Python type system.
    def __init__(self, error: Any):
        self.error = error
        message = "[{cls}] {message}".format(
            cls=error.__class__.__name__,
            message=error.message if hasattr(error, "message") else None,
        )
        super().__init__(message)


def pipeline_selector_from_graphql(data: Mapping[str, Any]) -> JobSubsetSelector:
    asset_selection = cast("Iterable[dict[str, list[str]]] | None", data.get("assetSelection"))
    asset_check_selection = cast("Iterable[dict[str, Any]] | None", data.get("assetCheckSelection"))
    return JobSubsetSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        job_name=data.get("pipelineName") or data.get("jobName"),  # type: ignore
        op_selection=data.get("solidSelection"),
        asset_selection=(
            [AssetKey.from_graphql_input(asset_key) for asset_key in asset_selection]
            if asset_selection
            else None
        ),
        asset_check_selection=(
            [AssetCheckKey.from_graphql_input(asset_check) for asset_check in asset_check_selection]
            if asset_check_selection is not None
            else None
        ),
    )


def graph_selector_from_graphql(data: Mapping[str, Any]) -> GraphSelector:
    return GraphSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        graph_name=data["graphName"],
    )


class ExecutionParams(
    NamedTuple(
        "_ExecutionParams",
        [
            ("selector", JobSubsetSelector),
            ("run_config", Mapping[str, object]),
            ("mode", str | None),
            ("execution_metadata", "ExecutionMetadata"),
            ("step_keys", Sequence[str] | None),
        ],
    )
):
    def __new__(
        cls,
        selector: JobSubsetSelector,
        run_config: Mapping[str, object] | None,
        mode: str | None,
        execution_metadata: "ExecutionMetadata",
        step_keys: Sequence[str] | None,
    ):
        check.opt_list_param(step_keys, "step_keys", of_type=str)

        return super().__new__(
            cls,
            selector=check.inst_param(selector, "selector", JobSubsetSelector),
            run_config=check.opt_mapping_param(run_config, "run_config", key_type=str),
            mode=check.opt_str_param(mode, "mode"),
            execution_metadata=check.inst_param(
                execution_metadata, "execution_metadata", ExecutionMetadata
            ),
            step_keys=step_keys,
        )

    def to_graphql_input(self) -> Mapping[str, Any]:
        return {
            "selector": self.selector.to_graphql_input(),
            "runConfigData": self.run_config,
            "mode": self.mode,
            "executionMetadata": self.execution_metadata.to_graphql_input(),
            "stepKeys": self.step_keys,
        }


class ExecutionMetadata(
    NamedTuple(
        "_ExecutionMetadata",
        [
            ("run_id", str | None),
            ("tags", Mapping[str, str]),
            ("root_run_id", str | None),
            ("parent_run_id", str | None),
        ],
    )
):
    def __new__(
        cls,
        run_id: str | None,
        tags: Mapping[str, str],
        root_run_id: str | None = None,
        parent_run_id: str | None = None,
    ):
        return super().__new__(
            cls,
            check.opt_str_param(run_id, "run_id"),
            check.dict_param(tags, "tags", key_type=str, value_type=str),
            check.opt_str_param(root_run_id, "root_run_id"),
            check.opt_str_param(parent_run_id, "parent_run_id"),
        )

    def to_graphql_input(self) -> Mapping[str, Any]:
        return {
            "runId": self.run_id,
            "tags": [{"key": k, "value": v} for k, v in self.tags.items()],
            "rootRunId": self.root_run_id,
            "parentRunId": self.parent_run_id,
        }


def apply_cursor_limit_reverse(
    items: Sequence[str],
    cursor: str | None,
    limit: int | None,
    reverse: bool | None,
) -> Sequence[str]:
    start = 0
    end = len(items)
    index = 0

    if cursor:
        index = next(idx for (idx, item) in enumerate(items) if item == cursor)

        if reverse:
            end = index
        else:
            start = index + 1

    if limit:
        if reverse:
            start = end - limit
        else:
            end = start + limit

    return items[max(start, 0) : end]


def get_query_limit_with_default(provided_limit: int | None, default_limit: int) -> int:
    check.opt_int_param(provided_limit, "provided_limit")

    if provided_limit is None:
        return default_limit

    if provided_limit > default_limit:
        raise DagsterError(f"Limit of {provided_limit} is too large. Max is {default_limit}")

    return provided_limit


BackfillParams: TypeAlias = Mapping[str, Any]
AssetBackfillPreviewParams: TypeAlias = Mapping[str, Any]



from datetime import timedelta

from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.freshness import (
    FreshnessPolicy,
    FreshnessState,
    FreshnessStateChange,
)
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.test_utils import instance_for_test
from dagster._time import get_current_timestamp
from dagster_graphql.test.utils import define_out_of_process_context, execute_dagster_graphql

GET_INTERNAL_FRESHNESS_POLICY = """
query GetInternalFreshnessPolicy($assetKey: AssetKeyInput!) {
    assetNodes(assetKeys: [$assetKey]) {
        internalFreshnessPolicy {
            ... on TimeWindowFreshnessPolicy {
                failWindowSeconds
                warnWindowSeconds
            }
            ... on CronFreshnessPolicy {
                deadlineCron
                lowerBoundDeltaSeconds
                timezone
            }
        }
    }
}
"""

GET_FRESHNESS_STATUS_INFO = """
query GetFreshnessStatusInfo($assetKey: AssetKeyInput!) {
    assetNodes(assetKeys: [$assetKey]) {
        freshnessStatusInfo {
            freshnessStatus
            freshnessStatusMetadata {
                ... on AssetHealthFreshnessMeta {
                    lastMaterializedTimestamp
                }
            }
        }
    }
}
"""

GET_FRESHNESS_STATUS_CHANGED_TIMESTAMP = """
query GetFreshnessStatusChangedTimestamp($assetKey: AssetKeyInput!) {
    assetOrError(assetKey: $assetKey) {
        ... on Asset {
            freshnessStatusChangedTimestamp
        }
    }
}
"""


@asset(
    freshness_policy=FreshnessPolicy.time_window(
        fail_window=timedelta(minutes=10), warn_window=timedelta(minutes=5)
    )
)
def asset_with_freshness_with_warn_window():
    pass


@asset(freshness_policy=FreshnessPolicy.time_window(fail_window=timedelta(minutes=10)))
def asset_with_freshness():
    pass


def get_repo() -> RepositoryDefinition:
    return Definitions(
        assets=[
            asset_with_freshness,
            asset_with_freshness_with_warn_window,
        ]
    ).get_repository_def()


# There is a separate implementation for plus graphql tests.
def test_freshness():
    with instance_for_test() as instance:
        assert instance.internal_asset_freshness_enabled()
        with define_out_of_process_context(__file__, "get_repo", instance) as graphql_context:
            result = execute_dagster_graphql(
                graphql_context,
                GET_INTERNAL_FRESHNESS_POLICY,
                variables={"assetKey": asset_with_freshness.key.to_graphql_input()},
            )
            assert result.data["assetNodes"][0]["internalFreshnessPolicy"] is not None
            result = execute_dagster_graphql(
                graphql_context,
                """
query getFreshnessEnabled {
    instance {
        freshnessEvaluationEnabled
    }
}
                """,
                variables={},
            )
            assert result.data["instance"]["freshnessEvaluationEnabled"] is True

            # starts off with no status
            result = execute_dagster_graphql(
                graphql_context,
                GET_FRESHNESS_STATUS_INFO,
                variables={"assetKey": asset_with_freshness.key.to_graphql_input()},
            )
            assert result.data["assetNodes"][0]["freshnessStatusInfo"] is not None
            assert (
                result.data["assetNodes"][0]["freshnessStatusInfo"]["freshnessStatus"] == "UNKNOWN"
            )

            # freshnessStatusChangedTimestamp starts as None
            result = execute_dagster_graphql(
                graphql_context,
                GET_FRESHNESS_STATUS_CHANGED_TIMESTAMP,
                variables={"assetKey": asset_with_freshness.key.to_graphql_input()},
            )
            assert result.data["assetOrError"]["freshnessStatusChangedTimestamp"] is None

            # now it's healthy
            state_change_timestamp = get_current_timestamp()
            instance._report_runless_asset_event(  # noqa: SLF001
                asset_event=FreshnessStateChange(
                    key=asset_with_freshness.key,
                    previous_state=FreshnessState.UNKNOWN,
                    new_state=FreshnessState.PASS,
                    state_change_timestamp=state_change_timestamp,
                )
            )

            # make sure it comes out as healthy
            result = execute_dagster_graphql(
                graphql_context,
                GET_FRESHNESS_STATUS_INFO,
                variables={"assetKey": asset_with_freshness.key.to_graphql_input()},
            )
            assert result.data["assetNodes"][0]["freshnessStatusInfo"] is not None
            assert (
                result.data["assetNodes"][0]["freshnessStatusInfo"]["freshnessStatus"] == "HEALTHY"
            )

            # freshnessStatusChangedTimestamp should now have a value in milliseconds
            result = execute_dagster_graphql(
                graphql_context,
                GET_FRESHNESS_STATUS_CHANGED_TIMESTAMP,
                variables={"assetKey": asset_with_freshness.key.to_graphql_input()},
            )
            returned_ts = result.data["assetOrError"]["freshnessStatusChangedTimestamp"]
            assert returned_ts is not None



import dagster._check as check
import graphene
from dagster._core.remote_representation.external import RemoteExecutionPlan
from dagster._core.snap import ExecutionStepInputSnap, ExecutionStepOutputSnap, ExecutionStepSnap

from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.metadata import GrapheneMetadataItemDefinition
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneExecutionStepOutput(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)

    class Meta:
        name = "ExecutionStepOutput"

    def __init__(self, step_output_snap):
        super().__init__()
        self._step_output_snap = check.inst_param(
            step_output_snap, "step_output_snap", ExecutionStepOutputSnap
        )

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._step_output_snap.name


class GrapheneExecutionStepInput(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    dependsOn = non_null_list(lambda: GrapheneExecutionStep)

    class Meta:
        name = "ExecutionStepInput"

    def __init__(
        self, step_input_snap: ExecutionStepInputSnap, remote_execution_plan: RemoteExecutionPlan
    ):
        super().__init__()
        self._step_input_snap = check.inst_param(
            step_input_snap, "step_input_snap", ExecutionStepInputSnap
        )
        self._remote_execution_plan = check.inst_param(
            remote_execution_plan, "remote_execution_plan", RemoteExecutionPlan
        )

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._step_input_snap.name

    def resolve_dependsOn(self, _graphene_info: ResolveInfo):
        return [
            GrapheneExecutionStep(
                self._remote_execution_plan,
                self._remote_execution_plan.get_step_by_key(key),
            )
            # We filter at this layer to ensure that we do not return outputs that
            # do not exist in the execution plan
            for key in filter(
                self._remote_execution_plan.key_in_plan,
                self._step_input_snap.upstream_step_keys,
            )
        ]


class GrapheneStepKind(graphene.Enum):
    COMPUTE = "COMPUTE"
    UNRESOLVED_MAPPED = "UNRESOLVED_MAPPED"
    UNRESOLVED_COLLECT = "UNRESOLVED_COLLECT"

    class Meta:
        name = "StepKind"

    @property
    def description(self):
        if self == GrapheneStepKind.COMPUTE:
            return "This is a user-defined computation step"
        if self == GrapheneStepKind.UNRESOLVED_MAPPED:
            return "This is a mapped step that has not yet been resolved"
        if self == GrapheneStepKind.UNRESOLVED_COLLECT:
            return "This is a collect step that is not yet resolved"
        else:
            return None


class GrapheneExecutionStep(graphene.ObjectType):
    key = graphene.NonNull(graphene.String)
    inputs = non_null_list(GrapheneExecutionStepInput)
    outputs = non_null_list(GrapheneExecutionStepOutput)
    solidHandleID = graphene.NonNull(graphene.String)
    kind = graphene.NonNull(GrapheneStepKind)
    metadata = non_null_list(GrapheneMetadataItemDefinition)

    class Meta:
        name = "ExecutionStep"

    def __init__(
        self, remote_execution_plan: RemoteExecutionPlan, execution_step_snap: ExecutionStepSnap
    ):
        super().__init__()
        self._remote_execution_plan = check.inst_param(
            remote_execution_plan, "remote_execution_plan", RemoteExecutionPlan
        )
        self._plan_snapshot = remote_execution_plan.execution_plan_snapshot
        self._step_snap = check.inst_param(
            execution_step_snap, "execution_step_snap", ExecutionStepSnap
        )

    def resolve_metadata(self, _graphene_info: ResolveInfo):
        return [
            GrapheneMetadataItemDefinition(key=mdi.key, value=mdi.value)
            for mdi in self._step_snap.metadata_items
        ]

    def resolve_inputs(self, _graphene_info: ResolveInfo):
        return [
            GrapheneExecutionStepInput(inp, self._remote_execution_plan)
            for inp in self._step_snap.inputs
        ]

    def resolve_outputs(self, _graphene_info: ResolveInfo):
        return [GrapheneExecutionStepOutput(out) for out in self._step_snap.outputs]

    def resolve_key(self, _graphene_info: ResolveInfo):
        return self._step_snap.key

    def resolve_solidHandleID(self, _graphene_info: ResolveInfo):
        return self._step_snap.node_handle_id

    def resolve_kind(self, _graphene_info: ResolveInfo):
        return self._step_snap.kind.value


class GrapheneExecutionPlan(graphene.ObjectType):
    steps = non_null_list(GrapheneExecutionStep)
    artifactsPersisted = graphene.NonNull(graphene.Boolean)
    # TODO: remove this, as we should be using assetKeys instead
    assetSelection = non_null_list(graphene.String)
    assetKeys = non_null_list(GrapheneAssetKey)

    class Meta:
        name = "ExecutionPlan"

    def __init__(self, remote_execution_plan: RemoteExecutionPlan):
        super().__init__()
        self._remote_execution_plan = check.inst_param(
            remote_execution_plan, "remote_execution_plan", RemoteExecutionPlan
        )

    def resolve_steps(self, _graphene_info: ResolveInfo):
        return [
            GrapheneExecutionStep(
                self._remote_execution_plan,
                self._remote_execution_plan.get_step_by_key(step.key),
            )
            for step in self._remote_execution_plan.get_steps_in_plan()
        ]

    def resolve_artifactsPersisted(self, _graphene_info: ResolveInfo):
        return self._remote_execution_plan.execution_plan_snapshot.artifacts_persisted

    def resolve_assetSelection(self, _graphene_info: ResolveInfo):
        # TODO: remove this, as we should be using assetKeys instead
        return list(self._remote_execution_plan.execution_plan_snapshot.asset_selection)

    def resolve_assetKeys(self, _graphene_info: ResolveInfo) -> list[GrapheneAssetKey]:
        return [
            GrapheneAssetKey(path=asset_key.path)
            for asset_key in self._remote_execution_plan.execution_plan_snapshot.asset_selection
        ]


types = [
    GrapheneExecutionPlan,
    GrapheneExecutionStep,
    GrapheneExecutionStepInput,
    GrapheneExecutionStepOutput,
    GrapheneStepKind,
]



import copy
import tempfile
import time
from unittest import mock

import pytest
import yaml
from dagster import AssetMaterialization, Output, define_asset_job, job, op, repository
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.dependency import NodeHandle
from dagster._core.definitions.events import AssetObservation
from dagster._core.definitions.job_base import InMemoryJob
from dagster._core.events import DagsterEvent, DagsterEventType
from dagster._core.events.log import EventLogEntry
from dagster._core.execution.api import execute_run
from dagster._core.execution.plan.handle import StepHandle
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import PARENT_RUN_ID_TAG, ROOT_RUN_ID_TAG
from dagster._core.test_utils import instance_for_test
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._utils import Counter, traced_counter
from dagster_graphql.test.utils import (
    define_out_of_process_context,
    execute_dagster_graphql,
    infer_job_selector,
)

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
    ReadonlyGraphQLContextTestMatrix,
)

RUNS_QUERY = """
query PipelineRunsRootQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    ... on Pipeline {
      name
      pipelineSnapshotId
      runs {
        ...RunHistoryRunFragment
      }
    }
  }
}

fragment RunHistoryRunFragment on PipelineRun {
  runId
  status
  repositoryOrigin {
    repositoryLocationName
    repositoryName
    repositoryLocationMetadata {
      key
      value
    }
  }
  pipeline {
    ...on PipelineReference { name }
  }
  executionPlan {
    steps {
      key
    }
  }
  runConfig
  runConfigYaml
  mode
  canTerminate
  tags {
    key
    value
  }
}
"""

DELETE_RUN_MUTATION = """
mutation DeleteRun($runId: String!) {
  deletePipelineRun(runId: $runId) {
    __typename
    ... on DeletePipelineRunSuccess {
      runId
    }
    ... on PythonError {
      message
    }
  }
}
"""

ALL_TAG_KEYS_QUERY = """
{
  runTagKeysOrError {
    __typename
    ... on RunTagKeys {
      keys
    }
    ... on PythonError {
        message
        stack
    }
  }
}
"""

FILTERED_TAGS_QUERY = """
query FilteredRunTagsQuery($tagKeys: [String!]!) {
  runTagsOrError(tagKeys: $tagKeys) {
    ... on RunTags {
      tags {
        key
        values
      }
    }
  }
}
"""


ALL_RUNS_QUERY = """
{
  pipelineRunsOrError {
    ... on PipelineRuns {
      results {
        runId
        pipelineSnapshotId
        pipeline {
          __typename
          ... on PipelineReference {
            name
            solidSelection
          }
        }
      }
    }
  }
}
"""


FILTERED_RUN_QUERY = """
query PipelineRunsRootQuery($filter: RunsFilter!) {
  pipelineRunsOrError(filter: $filter) {
    ... on PipelineRuns {
      results {
        runId
        hasReExecutePermission
        hasTerminatePermission
        hasDeletePermission
      }
    }
  }
}
"""

RUN_IDS_QUERY = """
query RunIds($filter: RunsFilter, $cursor: String, $limit: Int) {
  runIdsOrError(filter: $filter, cursor: $cursor, limit: $limit) {
    ... on RunIds {
      results
    }
  }
}
"""

FILTERED_RUN_COUNT_QUERY = """
query PipelineRunsRootQuery($filter: RunsFilter!) {
  pipelineRunsOrError(filter: $filter) {
    ... on PipelineRuns {
      count
    }
  }
}
"""


RUN_GROUP_QUERY = """
query RunGroupQuery($runId: ID!) {
  runGroupOrError(runId: $runId) {
    ... on RunGroup {
      __typename
      rootRunId
      runs {
        runId
        pipeline {
          __typename
          ... on PipelineReference {
            name
          }
        }
      }
    }
    ... on RunGroupNotFoundError {
      __typename
      runId
      message
    }
  }
}
"""


REPOSITORY_RUNS_QUERY = """
query RepositoryRunsQuery($repositorySelector: RepositorySelector!) {
    repositoryOrError(repositorySelector: $repositorySelector) {
        ... on Repository {
            id
            name
            pipelines {
                id
                name
                runs(limit: 10) {
                    id
                    runId
                }
            }
        }
    }
}
"""

ASSET_RUNS_QUERY = """
query AssetRunsQuery($assetKey: AssetKeyInput!) {
    assetOrError(assetKey: $assetKey) {
        ... on Asset {
            assetMaterializations {
                label
                runOrError {
                    ... on PipelineRun {
                        id
                        runId
                    }
                }
            }
        }
    }
}
"""

RUN_ASSETS_QUERY = """
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on Run {
      assets {
        id
        key {
          path
        }
      }
    }
  }
}
"""


RUN_CONCURRENCY_QUERY = """
{
  pipelineRunsOrError {
    ... on PipelineRuns {
      results {
        runId
        status
        hasConcurrencyKeySlots
        rootConcurrencyKeys
        allPools
      }
    }
  }
}
"""

RUN_LOGS_QUERY = """
query RunLogsQuery($afterCursor:String, $limit:Int, $runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on Run {
        eventConnection(limit: $limit, afterCursor: $afterCursor) {
          cursor
          hasMore
          events {
            ... on ExecutionStepFailureEvent {
              stepKey
              eventType
              errorSource
              error {
                message
                stack
                causes {

                  className
                  stack
                  message
                }
              }
            }
          }
        }
      }
    }
}
"""

RUN_METRICS_QUERY = """
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on Run {
        runId
        tags {
          key
          value
        }
        hasRunMetricsEnabled
    }
  }
}
"""

RUN_STEP_STATS_QUERY = """
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on Run {
        runId
        stepStats {
            stepKey
            status
        }
    }
  }
}
"""


def _get_runs_data(result, run_id):
    for run_data in result.data["pipelineOrError"]["runs"]:
        if run_data["runId"] == run_id:
            # so caller can delete keys
            return copy.deepcopy(run_data)

    return None


class TestDeleteRunReadonly(ReadonlyGraphQLContextTestMatrix):
    def test_delete_runs_permission_readonly(self, graphql_context: WorkspaceRequestContext):
        repo = get_repo_at_time_1()
        run_id = graphql_context.instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED.value
        ).run_id

        result = execute_dagster_graphql(
            graphql_context, DELETE_RUN_MUTATION, variables={"runId": run_id}
        )

        assert result.data["deletePipelineRun"]["__typename"] == "UnauthorizedError"

        result = execute_dagster_graphql(
            graphql_context,
            FILTERED_RUN_QUERY,
            variables={"filter": {"runIds": [run_id]}},
        )
        run_response = result.data["pipelineRunsOrError"]["results"][0]
        assert run_response["hasReExecutePermission"] is False
        assert run_response["hasTerminatePermission"] is False
        assert run_response["hasDeletePermission"] is False


class TestGetRuns(ExecutingGraphQLContextTestMatrix):
    def test_get_runs_over_graphql(self, graphql_context: WorkspaceRequestContext):
        # This include needs to be here because its inclusion screws up
        # other code in this file which reads itself to load a repo
        from dagster_graphql_tests.graphql.utils import sync_execute_get_run_log_data

        selector = infer_job_selector(graphql_context, "required_resource_job")

        payload_one = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "runConfigData": {"resources": {"R1": {"config": 2}}},
                    "executionMetadata": {"tags": [{"key": "fruit", "value": "apple"}]},
                }
            },
        )
        run_id_one = payload_one["run"]["runId"]

        read_context = graphql_context

        result = execute_dagster_graphql(read_context, RUNS_QUERY, variables={"selector": selector})

        runs = result.data["pipelineOrError"]["runs"]
        assert len(runs) == 1

        tags = runs[0]["tags"]

        tags_by_key = {tag["key"]: tag["value"] for tag in tags}

        assert tags_by_key["fruit"] == "apple"

        origin = runs[0]["repositoryOrigin"]
        assert origin
        assert origin["repositoryLocationName"] == selector["repositoryLocationName"]
        assert origin["repositoryName"] == selector["repositoryName"]

        payload_two = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "runConfigData": {"resources": {"R1": {"config": 3}}},
                    "executionMetadata": {"tags": [{"key": "veggie", "value": "carrot"}]},
                }
            },
        )

        run_id_two = payload_two["run"]["runId"]

        result = execute_dagster_graphql(read_context, RUNS_QUERY, variables={"selector": selector})

        runs = result.data["pipelineOrError"]["runs"]
        assert len(runs) == 2

        all_tag_keys_result = execute_dagster_graphql(read_context, ALL_TAG_KEYS_QUERY)
        tag_keys = set(all_tag_keys_result.data["runTagKeysOrError"]["keys"])
        # check presence rather than set equality since we might have extra tags in cloud
        assert "fruit" in tag_keys
        assert "veggie" in tag_keys

        filtered_tags_result = execute_dagster_graphql(
            read_context, FILTERED_TAGS_QUERY, variables={"tagKeys": ["fruit"]}
        )
        tags = filtered_tags_result.data["runTagsOrError"]["tags"]
        tags_dict = {item["key"]: item["values"] for item in tags}
        assert len(tags_dict) == 1
        assert tags_dict["fruit"] == ["apple"]

        run_logs_result = execute_dagster_graphql(
            read_context, RUN_LOGS_QUERY, variables={"runId": run_id_one, "limit": 5}
        )
        run_log_events = run_logs_result.data["pipelineRunOrError"]["eventConnection"]["events"]
        assert len(run_log_events) == 5
        assert run_logs_result.data["pipelineRunOrError"]["eventConnection"]["hasMore"]

        cursor = run_logs_result.data["pipelineRunOrError"]["eventConnection"]["cursor"]
        assert cursor

        run_logs_result = execute_dagster_graphql(
            read_context,
            RUN_LOGS_QUERY,
            variables={"runId": run_id_one, "limit": 1000, "afterCursor": cursor},
        )

        assert not run_logs_result.data["pipelineRunOrError"]["eventConnection"]["hasMore"]

        with pytest.raises(Exception, match=r"Limit of 5000 is too large. Max is 1000"):
            run_logs_result = execute_dagster_graphql(
                read_context,
                RUN_LOGS_QUERY,
                variables={"runId": run_id_one, "limit": 5000, "afterCursor": cursor},
            )

        with mock.patch.object(
            type(read_context),
            "records_for_run_default_limit",
            new_callable=mock.PropertyMock,
        ) as mock_records_for_run_default_limit:
            mock_records_for_run_default_limit.return_value = 5
            run_logs_result = execute_dagster_graphql(
                read_context,
                RUN_LOGS_QUERY,
                variables={"runId": run_id_one, "afterCursor": cursor},
            )
            assert len(run_logs_result.data["pipelineRunOrError"]["eventConnection"]["events"]) == 5

            with pytest.raises(Exception, match=r"Limit of 1000 is too large. Max is 5"):
                run_logs_result = execute_dagster_graphql(
                    read_context,
                    RUN_LOGS_QUERY,
                    variables={"runId": run_id_one, "limit": 1000, "afterCursor": cursor},
                )

            run_logs_result = execute_dagster_graphql(
                read_context,
                RUN_LOGS_QUERY,
                variables={"runId": run_id_one, "limit": 4, "afterCursor": cursor},
            )
            assert len(run_logs_result.data["pipelineRunOrError"]["eventConnection"]["events"]) == 4

        # delete the second run
        result = execute_dagster_graphql(
            read_context, DELETE_RUN_MUTATION, variables={"runId": run_id_two}
        )
        assert result.data["deletePipelineRun"]["__typename"] == "DeletePipelineRunSuccess"
        assert result.data["deletePipelineRun"]["runId"] == run_id_two

        # query it back out
        result = execute_dagster_graphql(read_context, RUNS_QUERY, variables={"selector": selector})

        # first is the same
        run_one_data = _get_runs_data(result, run_id_one)
        assert run_one_data

        # second is gone
        run_two_data = _get_runs_data(result, run_id_two)
        assert run_two_data is None

        # try to delete the second run again
        execute_dagster_graphql(read_context, DELETE_RUN_MUTATION, variables={"runId": run_id_two})

        result = execute_dagster_graphql(
            read_context, DELETE_RUN_MUTATION, variables={"runId": run_id_two}
        )
        assert result.data["deletePipelineRun"]["__typename"] == "RunNotFoundError"

    def test_run_config(self, graphql_context: WorkspaceRequestContext):
        # This include needs to be here because its inclusion screws up
        # other code in this file which reads itself to load a repo
        from dagster_graphql_tests.graphql.utils import sync_execute_get_run_log_data

        selector = infer_job_selector(graphql_context, "required_resource_job")

        sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "runConfigData": {"resources": {"R1": {"config": 2}}},
                }
            },
        )
        result = execute_dagster_graphql(
            graphql_context, RUNS_QUERY, variables={"selector": selector}
        )

        runs = result.data["pipelineOrError"]["runs"]
        assert len(runs) == 1
        run = runs[0]
        assert run["runConfig"] == {"resources": {"R1": {"config": 2}}}
        assert run["runConfigYaml"] == yaml.safe_dump(
            run["runConfig"], default_flow_style=False, allow_unicode=True
        )


def get_repo_at_time_1():
    @op
    def op_A():
        pass

    @op
    def op_B():
        pass

    @job
    def evolving_job():
        op_A()
        op_B()

    @job
    def foo_job():
        op_A()

    @repository
    def evolving_repo():
        return [evolving_job, foo_job]

    return evolving_repo


def get_repo_at_time_2():
    @op
    def op_A():
        pass

    @op
    def op_B_prime():
        pass

    @job
    def evolving_job():
        op_A()
        op_B_prime()

    @job
    def bar_job():
        op_A()

    @repository
    def evolving_repo():
        return [evolving_job, bar_job]

    return evolving_repo


def get_asset_repo():
    @op(tags={"dagster/concurrency_key": "foo"})
    def foo():
        yield AssetMaterialization(asset_key="foo", description="foo")
        yield AssetObservation(asset_key="bar", description="bar")
        yield Output(None)

    @job
    def foo_job():
        foo()

    @asset
    def my_fail_asset():
        raise Exception("OOPS")

    my_fail_asset_job = define_asset_job(name="my_fail_asset_job", selection="my_fail_asset")

    @repository
    def asset_repo():
        return [foo_job, my_fail_asset, my_fail_asset_job]

    return asset_repo


def test_runs_over_time():
    with instance_for_test() as instance:
        repo_1 = get_repo_at_time_1()

        full_evolve_run_id = (
            repo_1.get_job("evolving_job").execute_in_process(instance=instance).run_id
        )
        foo_run_id = repo_1.get_job("foo_job").execute_in_process(instance=instance).run_id
        evolve_a_run_id = (
            repo_1.get_job("evolving_job")
            .get_subset(op_selection={"op_A"})
            .execute_in_process(
                instance=instance,
            )
            .run_id
        )
        evolve_b_run_id = (
            repo_1.get_job("evolving_job")
            .get_subset(op_selection={"op_B"})
            .execute_in_process(
                instance=instance,
            )
            .run_id
        )

        with define_out_of_process_context(
            __file__, "get_repo_at_time_1", instance
        ) as context_at_time_1:
            result = execute_dagster_graphql(context_at_time_1, ALL_RUNS_QUERY)
            assert result.data

            t1_runs = {run["runId"]: run for run in result.data["pipelineRunsOrError"]["results"]}

            assert t1_runs[full_evolve_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }

            assert t1_runs[foo_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "foo_job",
                "solidSelection": None,
            }

            assert t1_runs[evolve_a_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }

            assert t1_runs[evolve_b_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }

        with define_out_of_process_context(
            __file__, "get_repo_at_time_2", instance
        ) as context_at_time_2:
            result = execute_dagster_graphql(context_at_time_2, ALL_RUNS_QUERY)
            assert result.data

            t2_runs = {run["runId"]: run for run in result.data["pipelineRunsOrError"]["results"]}

            assert t2_runs[full_evolve_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }

            assert t2_runs[evolve_a_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }
            # pipeline name changed
            assert t2_runs[foo_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "foo_job",
                "solidSelection": None,
            }
            # subset no longer valid - b renamed
            assert t2_runs[evolve_b_run_id]["pipeline"] == {
                "__typename": "PipelineSnapshot",
                "name": "evolving_job",
                "solidSelection": None,
            }


def test_filtered_runs():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        run_id_1 = (
            repo.get_job("foo_job")
            .execute_in_process(instance=instance, tags={"run": "one"})
            .run_id
        )
        run_id_2 = (
            repo.get_job("foo_job")
            .execute_in_process(instance=instance, tags={"run": "two"})
            .run_id
        )
        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={"filter": {"runIds": [run_id_1]}},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert run_ids[0] == run_id_1

            run_response = result.data["pipelineRunsOrError"]["results"][0]
            assert run_response["hasReExecutePermission"] is True
            assert run_response["hasTerminatePermission"] is True
            assert run_response["hasDeletePermission"] is True

            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={"filter": {"tags": [{"key": "run", "value": "one"}]}},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert run_ids[0] == run_id_1

            # test multiple run ids
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={"filter": {"runIds": [run_id_1, run_id_2]}},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 2
            assert set(run_ids) == set([run_id_1, run_id_2])


def test_filtered_runs_status():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        _ = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
        ).run_id
        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={"filter": {"statuses": ["FAILURE"]}},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert run_ids[0] == run_id_2


def test_filtered_runs_multiple_statuses():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        _ = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
        ).run_id
        run_id_3 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.SUCCESS
        ).run_id
        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={"filter": {"statuses": ["FAILURE", "SUCCESS"]}},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 2
            assert run_id_2 in run_ids
            assert run_id_3 in run_ids


def test_filtered_runs_multiple_filters():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()

        started_run_with_tags = instance.create_run_for_job(
            repo.get_job("foo_job"),
            status=DagsterRunStatus.STARTED,
            tags={"foo": "bar"},
        )
        failed_run_with_tags = instance.create_run_for_job(
            repo.get_job("foo_job"),
            status=DagsterRunStatus.FAILURE,
            tags={"foo": "bar"},
        )
        started_run_without_tags = instance.create_run_for_job(
            repo.get_job("foo_job"),
            status=DagsterRunStatus.STARTED,
            tags={"baz": "boom"},
        )

        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_QUERY,
                variables={
                    "filter": {
                        "statuses": ["STARTED"],
                        "tags": [{"key": "foo", "value": "bar"}],
                    }
                },
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert started_run_with_tags.run_id in run_ids
            assert failed_run_with_tags.run_id not in run_ids
            assert started_run_without_tags.run_id not in run_ids


def test_filtered_runs_count():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        instance.create_run_for_job(  # noqa: B018
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED
        ).run_id
        instance.create_run_for_job(  # noqa: B018
            repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
        ).run_id
        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                FILTERED_RUN_COUNT_QUERY,
                variables={"filter": {"statuses": ["FAILURE"]}},
            )
            assert result.data
            count = result.data["pipelineRunsOrError"]["count"]
            assert count == 1


def test_run_ids():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        run_id_1 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
        ).run_id
        run_id_3 = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
        ).run_id
        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                RUN_IDS_QUERY,
                variables={"filter": {"statuses": ["FAILURE"]}},
            )
            assert result.data
            assert result.data["runIdsOrError"]["results"] == [run_id_3, run_id_2]

            result = execute_dagster_graphql(
                context,
                RUN_IDS_QUERY,
                variables={"filter": {"statuses": ["FAILURE"]}, "limit": 1},
            )
            assert result.data
            assert result.data["runIdsOrError"]["results"] == [run_id_3]

            result = execute_dagster_graphql(
                context,
                RUN_IDS_QUERY,
                variables={"cursor": run_id_2},
            )
            assert result.data
            assert result.data["runIdsOrError"]["results"] == [run_id_1]


def test_run_group():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        foo_job = repo.get_job("foo_job")
        runs = [foo_job.execute_in_process(instance=instance)]
        root_run_id = runs[-1].run_id
        for _ in range(3):
            # https://github.com/dagster-io/dagster/issues/2433
            run = instance.create_run_for_job(
                foo_job,
                parent_run_id=root_run_id,
                root_run_id=root_run_id,
                tags={PARENT_RUN_ID_TAG: root_run_id, ROOT_RUN_ID_TAG: root_run_id},
            )
            execute_run(InMemoryJob(foo_job), run, instance)
            runs.append(run)  # pyright: ignore[reportArgumentType]

        with define_out_of_process_context(
            __file__, "get_repo_at_time_1", instance
        ) as context_at_time_1:
            result_one = execute_dagster_graphql(
                context_at_time_1,
                RUN_GROUP_QUERY,
                variables={"runId": root_run_id},
            )
            assert result_one.data["runGroupOrError"]["__typename"] == "RunGroup"

            assert len(result_one.data["runGroupOrError"]["runs"]) == 4

            result_two = execute_dagster_graphql(
                context_at_time_1,
                RUN_GROUP_QUERY,
                variables={"runId": runs[-1].run_id},
            )
            assert result_one.data["runGroupOrError"]["__typename"] == "RunGroup"
            assert len(result_two.data["runGroupOrError"]["runs"]) == 4

            assert (
                result_one.data["runGroupOrError"]["rootRunId"]
                == result_two.data["runGroupOrError"]["rootRunId"]
            )
            assert (
                result_one.data["runGroupOrError"]["runs"]
                == result_two.data["runGroupOrError"]["runs"]
            )


def test_run_group_not_found():
    with instance_for_test() as instance:
        with define_out_of_process_context(
            __file__, "get_repo_at_time_1", instance
        ) as context_at_time_1:
            result = execute_dagster_graphql(
                context_at_time_1,
                RUN_GROUP_QUERY,
                variables={"runId": "foo"},
            )
            assert result.data
            assert result.data["runGroupOrError"]
            assert result.data["runGroupOrError"]["__typename"] == "RunGroupNotFoundError"
            assert result.data["runGroupOrError"]["runId"] == "foo"
            assert result.data["runGroupOrError"][
                "message"
            ] == "Run group of run {run_id} could not be found.".format(run_id="foo")


def test_asset_batching():
    with instance_for_test() as instance:
        repo = get_asset_repo()
        foo_job = repo.get_job("foo_job")
        for _ in range(3):
            foo_job.execute_in_process(instance=instance)

        my_fail_asset_job = repo.get_job("my_fail_asset_job")
        result = my_fail_asset_job.execute_in_process(instance=instance, raise_on_error=False)

        fail_run_id = result.dagster_run.run_id

        with define_out_of_process_context(__file__, "asset_repo", instance) as context:
            traced_counter.set(Counter())
            result = execute_dagster_graphql(
                context, ASSET_RUNS_QUERY, variables={"assetKey": {"path": ["foo"]}}
            )
            assert result.data
            assert "assetOrError" in result.data
            assert "assetMaterializations" in result.data["assetOrError"]
            materializations = result.data["assetOrError"]["assetMaterializations"]
            assert len(materializations) == 3

            counter = traced_counter.get()
            counts = counter.counts()  # pyright: ignore[reportOptionalMemberAccess]
            assert counts
            assert counts.get("DagsterInstance.get_run_records") == 1

            run_ids = [materialization["runOrError"]["id"] for materialization in materializations]

            run_id = run_ids[0]
            result = execute_dagster_graphql(context, RUN_ASSETS_QUERY, variables={"runId": run_id})

            asset_ids = [asset["id"] for asset in result.data["pipelineRunOrError"]["assets"]]
            assert sorted(asset_ids) == sorted(['["foo"]', '["bar"]'])

            fail_run_result = execute_dagster_graphql(
                context, RUN_ASSETS_QUERY, variables={"runId": fail_run_id}
            )
            # despite no materialization or observation, the asset is returned, because it was planned
            asset_ids = [
                asset["id"] for asset in fail_run_result.data["pipelineRunOrError"]["assets"]
            ]
            assert asset_ids == ['["my_fail_asset"]']


def test_run_has_concurrency_slots():
    with tempfile.TemporaryDirectory() as temp_dir:
        with instance_for_test(
            overrides={
                "event_log_storage": {
                    "module": "dagster.utils.test",
                    "class": "ConcurrencyEnabledSqliteTestEventLogStorage",
                    "config": {"base_dir": temp_dir},
                },
            }
        ) as instance:
            repo = get_asset_repo()
            instance.event_log_storage.set_concurrency_slots("foo", 1)
            run_id = instance.create_run_for_job(
                repo.get_job("foo_job"), status=DagsterRunStatus.FAILURE
            ).run_id

            with define_out_of_process_context(__file__, "asset_repo", instance) as context:
                result = execute_dagster_graphql(context, RUN_CONCURRENCY_QUERY)
                assert result.data
                assert len(result.data["pipelineRunsOrError"]["results"]) == 1
                assert result.data["pipelineRunsOrError"]["results"][0]["runId"] == run_id
                assert not result.data["pipelineRunsOrError"]["results"][0][
                    "hasConcurrencyKeySlots"
                ]
                assert result.data["pipelineRunsOrError"]["results"][0]["rootConcurrencyKeys"]
                assert result.data["pipelineRunsOrError"]["results"][0]["allPools"]

            claim = instance.event_log_storage.claim_concurrency_slot(
                "foo", run_id, "fake_step_key"
            )
            assert claim.is_claimed

            with define_out_of_process_context(__file__, "asset_repo", instance) as context:
                result = execute_dagster_graphql(context, RUN_CONCURRENCY_QUERY)
                assert result.data
                assert len(result.data["pipelineRunsOrError"]["results"]) == 1
                assert result.data["pipelineRunsOrError"]["results"][0]["runId"] == run_id
                assert result.data["pipelineRunsOrError"]["results"][0]["hasConcurrencyKeySlots"]
                assert result.data["pipelineRunsOrError"]["results"][0]["rootConcurrencyKeys"]
                assert result.data["pipelineRunsOrError"]["results"][0]["allPools"]


@pytest.mark.parametrize(
    "run_tag, run_tag_value, run_metrics_enabled, failure_message",
    [
        (None, None, False, "run_metrics tag not present should result to false"),
        (".dagster/run_metrics", "true", True, "run_metrics tag set to true should result to true"),
        (
            ".dagster/run_metrics",
            "false",
            False,
            "run_metrics tag set to falsy value should result to false",
        ),
        (
            "dagster/run_metrics",
            "true",
            True,
            "public run_metrics tag set to true should result to true",
        ),
    ],
)
def test_run_has_run_metrics_enabled(run_tag, run_tag_value, run_metrics_enabled, failure_message):
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        tags = (
            {
                run_tag: run_tag_value,
            }
            if run_tag
            else {}
        )
        run = instance.create_run_for_job(
            repo.get_job("foo_job"), status=DagsterRunStatus.STARTED, tags=tags
        )

        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                RUN_METRICS_QUERY,
                variables={"runId": run.run_id},
            )
            assert result.data
            has_run_metrics_enabled = result.data["pipelineRunOrError"]["hasRunMetricsEnabled"]
            assert has_run_metrics_enabled == run_metrics_enabled, failure_message


def test_event_log_step_stats_retry_with_no_start():
    with instance_for_test() as instance:
        repo = get_repo_at_time_1()
        run = instance.create_run_for_job(repo.get_job("foo_job"), status=DagsterRunStatus.STARTED)
        storage = instance.event_log_storage
        node_handle = NodeHandle("E", None)
        step_handle = StepHandle(node_handle)
        storage.store_event(
            EventLogEntry(
                error_info=None,
                user_message="",
                level="debug",
                run_id=run.run_id,
                timestamp=time.time() - 150,
                step_key=step_handle.to_key(),
                job_name="foo_job",
                dagster_event=DagsterEvent(
                    DagsterEventType.STEP_UP_FOR_RETRY.value,
                    "foo_job",
                    node_handle=node_handle,
                    step_handle=step_handle,
                    event_specific_data=None,
                ),
            )
        )

        with define_out_of_process_context(__file__, "get_repo_at_time_1", instance) as context:
            result = execute_dagster_graphql(
                context,
                RUN_STEP_STATS_QUERY,
                variables={"runId": run.run_id},
            )
            assert result.data
            step_stats = result.data["pipelineRunOrError"]["stepStats"]
            assert len(step_stats) == 1
            step_stat = step_stats[0]
            assert step_stat["stepKey"] == "E"
            assert step_stat["status"] is None



import time

from dagster import AssetKey, DagsterEvent, DagsterEventType
from dagster._core.definitions.asset_checks.asset_check_evaluation import (
    AssetCheckEvaluation,
    AssetCheckEvaluationPlanned,
    AssetCheckEvaluationTargetMaterializationData,
)
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset
from dagster._core.event_api import EventLogRecord
from dagster._core.events import StepMaterializationData
from dagster._core.events.log import EventLogEntry
from dagster._core.test_utils import create_run_for_test, poll_for_finished_run
from dagster._core.utils import make_new_run_id
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.client.query import ERROR_FRAGMENT
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)

GET_ASSET_CHECK_DETAILS_QUERY = """
query AssetNodeChecksLimitQuery($assetKeys: [AssetKeyInput!], $limit: Int) {
    assetNodes(assetKeys: $assetKeys) {
        assetKey {
            path
        }
        assetChecksOrError(limit: $limit) {
            ... on AssetChecks {
                checks {
                    name
                    description
                    canExecuteIndividually
                    blocking
                    additionalAssetKeys {
                        path
                    }
                }
            }
        }
    }
}
"""

GET_ASSET_CHECK_NAMES_QUERY = """
query AssetNodeChecksLimitQuery($assetKeys: [AssetKeyInput!], $limit: Int, $pipelineSelector: PipelineSelector) {
    assetNodes(assetKeys: $assetKeys, pipeline: $pipelineSelector) {
        assetKey {
            path
        }
        assetChecksOrError(limit: $limit, pipeline: $pipelineSelector) {
            ... on AssetChecks {
                checks {
                    name
                }
            }
        }
    }
}
"""

GET_ASSET_CHECK_HISTORY_WITH_ID = """
query GetAssetChecksQuery($assetKey: AssetKeyInput!, $checkName: String!, $limit: Int!, $cursor: String) {
    assetCheckExecutions(assetKey: $assetKey, checkName: $checkName, limit: $limit, cursor: $cursor) {
        id
        status
        runId
    }
}
"""


GET_ASSET_CHECK_HISTORY = """
query GetAssetChecksQuery($assetKey: AssetKeyInput!, $checkName: String!) {
    assetCheckExecutions(assetKey: $assetKey, checkName: $checkName, limit: 10) {
        runId
        status
        evaluation {
            severity
            timestamp
            targetMaterialization {
                storageId
                runId
                timestamp
            }
            metadataEntries {
                label
            }
            description
        }
    }
}
"""

GET_LATEST_EXECUTION = """
query GetLatestExecution($assetKey: AssetKeyInput!) {
    assetNodes(assetKeys: [$assetKey]) {
        assetChecksOrError {
            ... on AssetChecks {
                checks {
                    name
                    executionForLatestMaterialization {
                        runId
                        status
                    }
                }
            }
        }
    }
}
"""

GET_ASSET_CHECK_HISTORY_WITH_STEP_KEY = """
query GetAssetChecksQuery($assetKey: AssetKeyInput!, $checkName: String!) {
    assetCheckExecutions(assetKey: $assetKey, checkName: $checkName, limit: 10) {
        runId
        status
        stepKey
    }
}
"""

GET_LOGS_FOR_RUN = """
query GetLogsForRun($runId: ID!) {
  logsForRun(runId: $runId) {
    __typename
    ... on EventConnection {
      events {
        __typename
      }
    }
  }
}
"""


LAUNCH_PIPELINE_EXECUTION_MUTATION = (
    ERROR_FRAGMENT
    + """
mutation($executionParams: ExecutionParams!) {
  launchPipelineExecution(executionParams: $executionParams) {
    __typename

    ... on InvalidStepError {
      invalidStepKey
    }
    ... on InvalidOutputError {
      stepKey
      invalidOutputName
    }
    ... on LaunchRunSuccess {
      run {
        runId
        pipeline {
          name
        }
        tags {
          key
          value
        }
        status
        runConfigYaml
        mode
        resolvedOpSelection
      }
    }
    ... on ConflictingExecutionParamsError {
      message
    }
    ... on PresetNotFoundError {
      preset
      message
    }
    ... on RunConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PythonError {
      ...errorFragment
    }
    ... on InvalidSubsetError {
        message
    }
  }
}
"""
)

RUN_QUERY = """
query RunQuery($runId: ID!) {
  runOrError(runId: $runId) {
    __typename
    ... on Run {
        assetSelection {
            path
        }
        assetCheckSelection {
            assetKey {
                path
            }
            name
        }
      }
    }
  }
"""

RUN_ASSET_CHECKS_QUERY = """
query RunAssetChecksQuery($runId: ID!) {
    pipelineRunOrError(runId: $runId) {
        ... on Run {
            id
            assetChecks {
                name
                assetKey {
                    path
                }
            }
        }
    }
}
"""


def _planned_event(run_id: str, planned: AssetCheckEvaluationPlanned) -> EventLogEntry:
    return EventLogEntry(
        error_info=None,
        user_message="",
        level="debug",
        run_id=run_id,
        timestamp=time.time(),
        dagster_event=DagsterEvent(
            DagsterEventType.ASSET_CHECK_EVALUATION_PLANNED.value,
            "nonce",
            event_specific_data=planned,
        ),
    )


def _evaluation_event(
    run_id: str, evaluation: AssetCheckEvaluation, timestamp=None
) -> EventLogEntry:
    return EventLogEntry(
        error_info=None,
        user_message="",
        level="debug",
        run_id=run_id,
        timestamp=timestamp or time.time(),
        dagster_event=DagsterEvent(
            DagsterEventType.ASSET_CHECK_EVALUATION.value,
            "nonce",
            event_specific_data=evaluation,
        ),
    )


def _materialization_event(run_id: str, asset_key: AssetKey) -> EventLogEntry:
    return EventLogEntry(
        error_info=None,
        user_message="",
        level="debug",
        run_id=run_id,
        timestamp=time.time(),
        dagster_event=DagsterEvent(
            DagsterEventType.ASSET_MATERIALIZATION.value,
            "nonce",
            event_specific_data=StepMaterializationData(
                materialization=AssetMaterialization(asset_key=asset_key), asset_lineage=[]
            ),
        ),
    )


class TestAssetChecks(ExecutingGraphQLContextTestMatrix):
    def test_asset_check_definitions(self, graphql_context: WorkspaceRequestContext):
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_DETAILS_QUERY,
            variables={"assetKeys": [{"path": ["asset_1"]}]},
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetKey": {"path": ["asset_1"]},
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "description": "asset_1 check",
                                "canExecuteIndividually": "CAN_EXECUTE",
                                "blocking": True,
                                "additionalAssetKeys": [
                                    {"path": ["asset_two"]},
                                ],
                            }
                        ]
                    },
                }
            ]
        }

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_DETAILS_QUERY,
            variables={"assetKeys": {"path": ["check_in_op_asset"]}},
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetKey": {"path": ["check_in_op_asset"]},
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "description": None,
                                "canExecuteIndividually": "REQUIRES_MATERIALIZATION",
                                "blocking": False,
                                "additionalAssetKeys": [],
                            }
                        ]
                    },
                }
            ]
        }

    def test_asset_check_asset_node_limit(self, graphql_context: WorkspaceRequestContext):
        res = execute_dagster_graphql(graphql_context, GET_ASSET_CHECK_NAMES_QUERY)
        with_checks = [
            node for node in res.data["assetNodes"] if node["assetChecksOrError"] != {"checks": []}
        ]
        assert with_checks == [
            {
                "assetKey": {"path": ["asset_1"]},
                "assetChecksOrError": {"checks": [{"name": "my_check"}]},
            },
            {
                "assetKey": {"path": ["asset_3"]},
                "assetChecksOrError": {
                    "checks": [{"name": "asset_3_check"}, {"name": "asset_3_other_check"}]
                },
            },
            {
                "assetKey": {"path": ["check_in_op_asset"]},
                "assetChecksOrError": {"checks": [{"name": "my_check"}]},
            },
            {
                "assetKey": {"path": ["one"]},
                "assetChecksOrError": {
                    "checks": [{"name": "my_check"}, {"name": "my_other_check"}]
                },
            },
            {
                "assetKey": {"path": ["owned_asset"]},
                "assetChecksOrError": {"checks": [{"name": "owned_asset_check"}]},
            },
            {
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "assetChecksOrError": {"checks": [{"name": "partitioned_asset_check"}]},
            },
            {
                "assetKey": {"path": ["unowned_asset"]},
                "assetChecksOrError": {"checks": [{"name": "unowned_asset_check"}]},
            },
        ]

        res = execute_dagster_graphql(
            graphql_context, GET_ASSET_CHECK_NAMES_QUERY, variables={"limit": 1}
        )
        with_checks = [
            node for node in res.data["assetNodes"] if node["assetChecksOrError"] != {"checks": []}
        ]
        assert with_checks == [
            {
                "assetKey": {"path": ["asset_1"]},
                "assetChecksOrError": {"checks": [{"name": "my_check"}]},
            },
            {
                "assetKey": {"path": ["asset_3"]},
                "assetChecksOrError": {"checks": [{"name": "asset_3_check"}]},
            },
            {
                "assetKey": {"path": ["check_in_op_asset"]},
                "assetChecksOrError": {"checks": [{"name": "my_check"}]},
            },
            {
                "assetKey": {"path": ["one"]},
                "assetChecksOrError": {"checks": [{"name": "my_check"}]},
            },
            {
                "assetKey": {"path": ["owned_asset"]},
                "assetChecksOrError": {"checks": [{"name": "owned_asset_check"}]},
            },
            {
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "assetChecksOrError": {"checks": [{"name": "partitioned_asset_check"}]},
            },
            {
                "assetKey": {"path": ["unowned_asset"]},
                "assetChecksOrError": {"checks": [{"name": "unowned_asset_check"}]},
            },
        ]

    def test_asset_check_asset_node_job_selection(self, graphql_context: WorkspaceRequestContext):
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_NAMES_QUERY,
            variables={
                "pipelineSelector": infer_job_selector(
                    graphql_context, "fail_partition_materialization_job"
                )
            },
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetKey": {"path": ["fail_partition_materialization"]},
                    "assetChecksOrError": {"checks": []},
                }
            ]
        }

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_NAMES_QUERY,
            variables={"pipelineSelector": infer_job_selector(graphql_context, "asset_check_job")},
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {"checks": [{"name": "my_check"}]},
                    "assetKey": {"path": ["asset_1"]},
                },
                {
                    "assetChecksOrError": {"checks": [{"name": "my_check"}]},
                    "assetKey": {"path": ["check_in_op_asset"]},
                },
            ]
        }

    def test_asset_check_execution_cursoring(self, graphql_context: WorkspaceRequestContext):
        run_ids = [create_run_for_test(graphql_context.instance).run_id for _ in range(10)]
        for run_id in run_ids:
            graphql_context.instance.event_log_storage.store_event(
                _planned_event(
                    run_id,
                    AssetCheckEvaluationPlanned(
                        asset_key=AssetKey(["asset_1"]), check_name="my_check"
                    ),
                )
            )

        expected_limit = 6
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_ID,
            variables={
                "assetKey": {"path": ["asset_1"]},
                "checkName": "my_check",
                "limit": expected_limit,
            },
        )
        assert res.data
        expected_run_ids = list(reversed(run_ids))[:expected_limit]
        for expected_run_id, execution in zip(expected_run_ids, res.data["assetCheckExecutions"]):
            assert execution["runId"] == expected_run_id

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_ID,
            variables={
                "assetKey": {"path": ["asset_1"]},
                "checkName": "my_check",
                "limit": expected_limit,
                "cursor": res.data["assetCheckExecutions"][-1]["id"],
            },
        )
        assert res.data
        expected_run_ids = list(reversed(run_ids))[expected_limit:]
        for expected_run_id, execution in zip(expected_run_ids, res.data["assetCheckExecutions"]):
            assert execution["runId"] == expected_run_id

    def test_asset_check_executions(self, graphql_context: WorkspaceRequestContext):
        run_id_one, run_id_two = [make_new_run_id() for _ in range(2)]
        create_run_for_test(graphql_context.instance, run_id=run_id_one)

        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_one,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={"assetKey": {"path": ["asset_1"]}, "checkName": "my_check"},
        )
        assert res.data == {
            "assetCheckExecutions": [
                {
                    "runId": run_id_one,
                    "status": "IN_PROGRESS",
                    "evaluation": None,
                }
            ],
        }

        evaluation_timestamp = time.time()

        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_one,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["asset_1"]),
                    check_name="my_check",
                    passed=True,
                    metadata={"foo": MetadataValue.text("bar")},
                    target_materialization_data=AssetCheckEvaluationTargetMaterializationData(
                        storage_id=42, run_id=run_id_two, timestamp=3.3
                    ),
                    severity=AssetCheckSeverity.ERROR,
                    description="evaluation description",
                ),
                timestamp=evaluation_timestamp,
            )
        )

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={"assetKey": {"path": ["asset_1"]}, "checkName": "my_check"},
        )
        assert res.data == {
            "assetCheckExecutions": [
                {
                    "runId": run_id_one,
                    "status": "SUCCEEDED",
                    "evaluation": {
                        "timestamp": evaluation_timestamp,
                        "severity": "ERROR",
                        "targetMaterialization": {
                            "storageId": "42",
                            "runId": run_id_two,
                            "timestamp": 3.3,
                        },
                        "metadataEntries": [
                            {"label": "foo"},
                        ],
                        "description": "evaluation description",
                    },
                }
            ],
        }

    def test_asset_check_events(self, graphql_context: WorkspaceRequestContext):
        run_id_one, run_id_two = [make_new_run_id() for _ in range(2)]
        create_run_for_test(graphql_context.instance, run_id=run_id_one)

        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_one,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )
        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_one,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["asset_1"]),
                    check_name="my_check",
                    passed=True,
                    metadata={"foo": MetadataValue.text("bar")},
                    target_materialization_data=AssetCheckEvaluationTargetMaterializationData(
                        storage_id=42, run_id=run_id_two, timestamp=3.3
                    ),
                    severity=AssetCheckSeverity.ERROR,
                ),
            )
        )

        res = execute_dagster_graphql(
            graphql_context, GET_LOGS_FOR_RUN, variables={"runId": run_id_one}
        )
        assert res.data == {
            "logsForRun": {
                "__typename": "EventConnection",
                "events": [
                    {
                        "__typename": "AssetCheckEvaluationPlannedEvent",
                    },
                    {
                        "__typename": "AssetCheckEvaluationEvent",
                    },
                ],
            }
        }

    def test_asset_check_failure(self, graphql_context: WorkspaceRequestContext):
        run = create_run_for_test(graphql_context.instance)

        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run.run_id,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={"assetKey": {"path": ["asset_1"]}, "checkName": "my_check"},
        )
        assert res.data == {
            "assetCheckExecutions": [
                {
                    "runId": run.run_id,
                    "status": "IN_PROGRESS",
                    "evaluation": None,
                }
            ],
        }

        graphql_context.instance.report_run_failed(run)

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={"assetKey": {"path": ["asset_1"]}, "checkName": "my_check"},
        )
        assert res.data == {
            "assetCheckExecutions": [
                {"runId": run.run_id, "status": "EXECUTION_FAILED", "evaluation": None}
            ],
        }

    def test_latest_execution(self, graphql_context: WorkspaceRequestContext):
        instance = graphql_context.instance

        def new_materialization() -> EventLogRecord:
            """Stores a materialization in a new run, and asserts that latest execustion is null."""
            run = create_run_for_test(instance)
            instance.event_log_storage.store_event(
                _materialization_event(run.run_id, AssetKey(["asset_1"]))
            )
            res = execute_dagster_graphql(
                graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
            )
            assert res.data == {
                "assetNodes": [
                    {
                        "assetChecksOrError": {
                            "checks": [
                                {"name": "my_check", "executionForLatestMaterialization": None}
                            ]
                        }
                    }
                ]
            }, "new materialization should clear latest execution"
            records = instance.get_asset_records([AssetKey(["asset_1"])])
            assert records
            assert records[0].asset_entry.last_materialization_record
            return records[0].asset_entry.last_materialization_record

        # no materialization, surface latest execution
        run = create_run_for_test(instance)
        instance.event_log_storage.store_event(
            _planned_event(
                run.run_id,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "IN_PROGRESS",
                                },
                            }
                        ]
                    }
                }
            ]
        }
        instance.report_run_failed(run)
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "EXECUTION_FAILED",
                                },
                            }
                        ]
                    },
                }
            ]
        }

        materialization_record = new_materialization()

        # check run targets the materialization
        run = create_run_for_test(instance)
        instance.event_log_storage.store_event(
            _planned_event(
                run.run_id,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "IN_PROGRESS",
                                },
                            }
                        ]
                    },
                }
            ]
        }
        instance.event_log_storage.store_event(
            _evaluation_event(
                run.run_id,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["asset_1"]),
                    check_name="my_check",
                    passed=True,
                    metadata={},
                    target_materialization_data=AssetCheckEvaluationTargetMaterializationData(
                        storage_id=materialization_record.storage_id,
                        run_id=materialization_record.event_log_entry.run_id,
                        timestamp=materialization_record.event_log_entry.timestamp,
                    ),
                    severity=AssetCheckSeverity.ERROR,
                ),
            )
        )
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "SUCCEEDED",
                                },
                            }
                        ]
                    },
                }
            ]
        }

        materialization_record = new_materialization()

        # run the checks, then materialize. The checks still apply because they're in the same run
        run = create_run_for_test(instance)
        instance.event_log_storage.store_event(
            _planned_event(
                run.run_id,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )
        instance.event_log_storage.store_event(
            _evaluation_event(
                run.run_id,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["asset_1"]),
                    check_name="my_check",
                    passed=True,
                    metadata={},
                    target_materialization_data=AssetCheckEvaluationTargetMaterializationData(
                        storage_id=materialization_record.storage_id,
                        run_id=materialization_record.event_log_entry.run_id,
                        timestamp=materialization_record.event_log_entry.timestamp,
                    ),
                    severity=AssetCheckSeverity.ERROR,
                ),
            )
        )
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "SUCCEEDED",
                                },
                            }
                        ]
                    },
                }
            ]
        }
        instance.event_log_storage.store_event(
            _materialization_event(run.run_id, AssetKey(["asset_1"]))
        )
        res = execute_dagster_graphql(
            graphql_context, GET_LATEST_EXECUTION, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {
            "assetNodes": [
                {
                    "assetChecksOrError": {
                        "checks": [
                            {
                                "name": "my_check",
                                "executionForLatestMaterialization": {
                                    "runId": run.run_id,
                                    "status": "SUCCEEDED",
                                },
                            }
                        ]
                    },
                }
            ]
        }

        new_materialization()

    def test_launch_subset_with_only_check(self, graphql_context: WorkspaceRequestContext):
        # materialize the asset and run the check first
        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
            asset_selection=[{"path": ["asset_1"]}],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess", (
            result.data
        )

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        result = execute_dagster_graphql(graphql_context, RUN_QUERY, variables={"runId": run_id})
        assert result.data == {
            "runOrError": {
                "__typename": "Run",
                "assetSelection": [{"path": ["asset_1"]}],
                "assetCheckSelection": None,
            }
        }

        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        checks = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(checks) == 1

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 1

        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
            asset_selection=[],
            asset_check_selection=[{"assetKey": {"path": ["asset_1"]}, "name": "my_check"}],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        result = execute_dagster_graphql(graphql_context, RUN_QUERY, variables={"runId": run_id})
        assert result.data == {
            "runOrError": {
                "__typename": "Run",
                "assetSelection": None,
                "assetCheckSelection": [
                    {"assetKey": {"path": ["asset_1"]}, "name": "my_check"},
                ],
            }
        }

        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 1

        for log in logs:
            if log.dagster_event:
                assert log.dagster_event.event_type != DagsterEventType.ASSET_MATERIALIZATION.value

    def test_launch_subset_asset_and_included_check(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
            asset_selection=[{"path": ["asset_1"]}],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        result = execute_dagster_graphql(graphql_context, RUN_QUERY, variables={"runId": run_id})
        assert result.data == {
            "runOrError": {
                "__typename": "Run",
                "assetSelection": [{"path": ["asset_1"]}],
                "assetCheckSelection": None,
            }
        }

        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 1

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 1

    def test_check_subset_error(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
            asset_check_selection=[
                {"assetKey": {"path": ["asset_1"]}, "name": "non-existent-check"}
            ],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "InvalidSubsetError"
        assert (
            result.data["launchPipelineExecution"]["message"]
            == "dagster._core.errors.DagsterInvalidSubsetError: AssetCheckKey(s)"
            " ['asset_1:non-existent-check'] were selected, but no definitions supply these keys."
            " Make sure all keys are spelled correctly, and all definitions are correctly added to"
            " the `Definitions`.\n"
        )

    def test_launch_subset_asset_no_check(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
            asset_selection=[{"path": ["asset_1"]}],
            asset_check_selection=[],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        for log in logs:
            if log.dagster_event:
                assert log.dagster_event.event_type != DagsterEventType.ASSET_CHECK_EVALUATION.value

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 1

    def test_multi_asset(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 2

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 2

    def test_multi_asset_without_check(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
            asset_selection=[{"path": ["one"]}],
            asset_check_selection=[],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 0

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 1

        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
            asset_selection=[{"path": ["two"]}],
            asset_check_selection=[],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 0

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 1

        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
            asset_selection=[{"path": ["one"]}, {"path": ["two"]}],
            asset_check_selection=[],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 0

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 2

    def test_multi_asset_only_check(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
            asset_selection=[],
            asset_check_selection=[
                {"assetKey": {"path": ["one"]}, "name": "my_check"},
            ],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id, timeout=10000000000)

        logs = graphql_context.instance.all_logs(run_id)
        assert run.is_success

        check_evaluations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_CHECK_EVALUATION
        ]
        assert len(check_evaluations) == 1

        materializations = [
            log
            for log in logs
            if log.dagster_event
            and log.dagster_event.event_type == DagsterEventType.ASSET_MATERIALIZATION
        ]
        assert len(materializations) == 0

    def test_step_key(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
            asset_selection=[],
            asset_check_selection=[
                {"assetKey": {"path": ["one"]}, "name": "my_check"},
            ],
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]
        run = poll_for_finished_run(graphql_context.instance, run_id, timeout=10000000000)

        assert run.is_success

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_STEP_KEY,
            variables={"assetKey": {"path": ["one"]}, "checkName": "my_check", "limit": 10},
        )
        assert result.data == {
            "assetCheckExecutions": [
                {
                    "runId": run_id,
                    "status": "SUCCEEDED",
                    "stepKey": "subsettable_checked_multi_asset",
                }
            ],
        }

    def test_deleted_run(self, graphql_context: WorkspaceRequestContext):
        run_id = make_new_run_id()
        run = create_run_for_test(graphql_context.instance, run_id=run_id)

        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id,
                AssetCheckEvaluationPlanned(asset_key=AssetKey(["asset_1"]), check_name="my_check"),
            )
        )

        graphql_context.instance.delete_run(run.run_id)

        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={"assetKey": {"path": ["asset_1"]}, "checkName": "my_check"},
        )
        assert res.data == {
            "assetCheckExecutions": [
                {
                    "runId": run_id,
                    "status": "SKIPPED",
                    "evaluation": None,
                }
            ],
        }

    def test_launch_asset_job_without_subset(self, graphql_context: WorkspaceRequestContext):
        # materialize the asset and run the check first
        selector = infer_job_selector(
            graphql_context,
            "asset_check_job",
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        result = execute_dagster_graphql(graphql_context, RUN_QUERY, variables={"runId": run_id})
        assert result.data == {
            "runOrError": {
                "__typename": "Run",
                "assetSelection": None,
                "assetCheckSelection": None,
            }
        }

        run = poll_for_finished_run(graphql_context.instance, run_id)

        assert run.is_success

        res = execute_dagster_graphql(
            graphql_context,
            RUN_ASSET_CHECKS_QUERY,
            variables={"runId": run_id},
        )

        assert res.data["pipelineRunOrError"]["assetChecks"] == [
            {"name": "my_check", "assetKey": {"path": ["asset_1"]}},
            {"name": "my_check", "assetKey": {"path": ["check_in_op_asset"]}},
        ]

    def test_launch_multi_asset_job_without_subset(self, graphql_context: WorkspaceRequestContext):
        # materialize the asset and run the check first
        selector = infer_job_selector(
            graphql_context,
            "checked_multi_asset_job",
        )
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                    "stepKeys": None,
                }
            },
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        result = execute_dagster_graphql(graphql_context, RUN_QUERY, variables={"runId": run_id})
        assert result.data == {
            "runOrError": {
                "__typename": "Run",
                "assetSelection": None,
                "assetCheckSelection": None,
            }
        }

        run = poll_for_finished_run(graphql_context.instance, run_id)

        assert run.is_success

        res = execute_dagster_graphql(
            graphql_context,
            RUN_ASSET_CHECKS_QUERY,
            variables={"runId": run_id},
        )

        assert res.data["pipelineRunOrError"]["assetChecks"] == [
            {"name": "my_check", "assetKey": {"path": ["one"]}},
            {"name": "my_other_check", "assetKey": {"path": ["one"]}},
        ]

    def test_partitioned_asset_check_executions(self, graphql_context: WorkspaceRequestContext):
        """Test retrieving asset check executions with partition subsets."""
        run_id_one, run_id_two, run_id_three = [make_new_run_id() for _ in range(3)]

        # Create runs for different partitions
        create_run_for_test(graphql_context.instance, run_id=run_id_one)
        create_run_for_test(graphql_context.instance, run_id=run_id_two)
        create_run_for_test(graphql_context.instance, run_id=run_id_three)

        # Store planned events for partition "a"
        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_one,
                AssetCheckEvaluationPlanned(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    partitions_subset=DefaultPartitionsSubset({"a"}),
                ),
            )
        )

        # Store planned events for partition "b"
        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_two,
                AssetCheckEvaluationPlanned(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    partitions_subset=DefaultPartitionsSubset({"b"}),
                ),
            )
        )

        # Query without partition filter - should get all executions
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "checkName": "partitioned_asset_check",
            },
        )

        executions = res.data["assetCheckExecutions"]
        assert len(executions) == 2
        run_ids = {execution["runId"] for execution in executions}
        assert run_ids == {run_id_one, run_id_two}

        # All should be in progress
        for execution in executions:
            assert execution["status"] == "IN_PROGRESS"

        # Store evaluation for partition "a" - passed
        evaluation_timestamp_a = time.time()
        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_one,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    passed=True,
                    metadata={"partition": MetadataValue.text("a")},
                    severity=AssetCheckSeverity.WARN,
                    description="Check passed for partition a",
                    partition="a",
                ),
                timestamp=evaluation_timestamp_a,
            )
        )

        # Store evaluation for partition "b" - failed
        evaluation_timestamp_b = time.time()
        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_two,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    passed=False,
                    metadata={"partition": MetadataValue.text("b")},
                    severity=AssetCheckSeverity.ERROR,
                    description="Check failed for partition b",
                    partition="b",
                ),
                timestamp=evaluation_timestamp_b,
            )
        )

        # Query again - should see completed evaluations
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY,
            variables={
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "checkName": "partitioned_asset_check",
            },
        )

        executions = res.data["assetCheckExecutions"]
        # Should have completed executions (with evaluation)
        completed_executions = [e for e in executions if e["evaluation"] is not None]
        assert len(completed_executions) == 2

        # Find partition a execution
        exec_a = next(e for e in completed_executions if e["runId"] == run_id_one)
        assert exec_a["status"] == "SUCCEEDED"
        assert exec_a["evaluation"]["severity"] == "WARN"
        assert exec_a["evaluation"]["description"] == "Check passed for partition a"

        # Find partition b execution
        exec_b = next(e for e in completed_executions if e["runId"] == run_id_two)
        assert exec_b["status"] == "FAILED"
        assert exec_b["evaluation"]["severity"] == "ERROR"
        assert exec_b["evaluation"]["description"] == "Check failed for partition b"

    def test_partitioned_asset_check_executions_with_partition_filter(
        self, graphql_context: WorkspaceRequestContext
    ):
        """Test retrieving asset check executions with partition filter."""
        # Define query with partition filter
        GET_ASSET_CHECK_HISTORY_WITH_PARTITION = """
        query GetAssetChecksQuery($assetKey: AssetKeyInput!, $checkName: String!, $partition: String) {
            assetCheckExecutions(assetKey: $assetKey, checkName: $checkName, limit: 10, partition: $partition) {
                runId
                status
                evaluation {
                    severity
                    description
                }
            }
        }
        """

        run_id_one, run_id_two = [make_new_run_id() for _ in range(2)]

        create_run_for_test(graphql_context.instance, run_id=run_id_one)
        create_run_for_test(graphql_context.instance, run_id=run_id_two)

        # Store evaluations for multiple partitions - using partitioned_asset_for_checks and partitioned_asset_check
        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_one,
                AssetCheckEvaluationPlanned(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    partitions_subset=DefaultPartitionsSubset({"2024-01"}),
                ),
            )
        )

        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_one,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    passed=True,
                    severity=AssetCheckSeverity.WARN,
                    description="Check for January",
                    partition="2024-01",
                ),
            )
        )

        graphql_context.instance.event_log_storage.store_event(
            _planned_event(
                run_id_two,
                AssetCheckEvaluationPlanned(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    partitions_subset=DefaultPartitionsSubset({"2024-02"}),
                ),
            )
        )

        graphql_context.instance.event_log_storage.store_event(
            _evaluation_event(
                run_id_two,
                AssetCheckEvaluation(
                    asset_key=AssetKey(["partitioned_asset_for_checks"]),
                    check_name="partitioned_asset_check",
                    passed=True,
                    severity=AssetCheckSeverity.WARN,
                    description="Check for February",
                    partition="2024-02",
                ),
            )
        )

        # Query with partition filter for "2024-01"
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_PARTITION,
            variables={
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "checkName": "partitioned_asset_check",
                "partition": "2024-01",
            },
        )

        executions = res.data["assetCheckExecutions"]
        assert len(executions) == 1
        assert executions[0]["runId"] == run_id_one
        assert executions[0]["evaluation"] is not None
        assert executions[0]["evaluation"]["description"] == "Check for January"

        # Query with partition filter for "2024-02"
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_PARTITION,
            variables={
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "checkName": "partitioned_asset_check",
                "partition": "2024-02",
            },
        )

        executions = res.data["assetCheckExecutions"]
        assert len(executions) == 1
        assert executions[0]["runId"] == run_id_two
        assert executions[0]["evaluation"] is not None
        assert executions[0]["evaluation"]["description"] == "Check for February"

        # Query without partition filter - should get both
        res = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_CHECK_HISTORY_WITH_PARTITION,
            variables={
                "assetKey": {"path": ["partitioned_asset_for_checks"]},
                "checkName": "partitioned_asset_check",
            },
        )

        executions = res.data["assetCheckExecutions"]
        assert len(executions) == 2



from dagster import ScheduleDefinition, job, op, repository


@op
def add_one(num: int) -> int:
    return num + 1


@op
def mult_two(num: int) -> int:
    return num * 2


@job
def math():
    mult_two(add_one())


@op(config_schema={"gimme": str})
def needs_config(context):
    return context.op_config["gimme"]


@op
def no_config():
    return "ok"


@job
def subset_test():
    no_config()
    needs_config()


def define_schedules():
    math_hourly_schedule = ScheduleDefinition(
        name="math_hourly_schedule",
        cron_schedule="0 0 * * *",
        job_name="math",
        run_config={"ops": {"add_one": {"inputs": {"num": {"value": 123}}}}},
    )

    return [math_hourly_schedule]


@repository
def the_test_repo():
    return [math, subset_test] + define_schedules()



"""Shared utilities for building partition status responses.

This module contains status-agnostic algorithms for building partition status
representations used by both regular assets and asset checks.
"""

from collections import defaultdict
from collections.abc import Callable, Mapping, Sequence
from enum import Enum
from typing import Any, Generic, TypeVar

from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import (
    MultiPartitionsDefinition,
    PartitionsDefinition,
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.subset import PartitionsSubset, TimeWindowPartitionsSubset
from dagster._core.instance import DynamicPartitionsStore
from dagster._record import record

StatusT = TypeVar("StatusT", bound=Enum)


@record
class TimePartitionRange(Generic[StatusT]):
    """Generic time partition range with status (status-agnostic)."""

    start_time: float
    end_time: float
    start_key: str
    end_key: str
    status: StatusT


def convert_time_window_subset_to_generic_ranges(
    subset: TimeWindowPartitionsSubset,
    status: StatusT,
) -> Sequence[TimePartitionRange[StatusT]]:
    """Convert a TimeWindowPartitionsSubset to generic range objects.

    Args:
        subset: Subset containing time windows
        status: Status to assign to all ranges

    Returns:
        List of TimePartitionRange objects with the given status
    """
    ranges = []
    partitions_def = subset.partitions_def

    for time_window in subset.included_time_windows:
        partition_key_range = partitions_def.get_partition_key_range_for_time_window(
            time_window.to_public_time_window()
        )
        ranges.append(
            TimePartitionRange(
                start_time=time_window.start_timestamp,
                end_time=time_window.end_timestamp,
                start_key=partition_key_range.start,
                end_key=partition_key_range.end,
                status=status,
            )
        )

    return ranges


def build_time_partition_ranges_generic(
    subsets: Mapping[StatusT, TimeWindowPartitionsSubset],
) -> Sequence[TimePartitionRange[StatusT]]:
    """Build time-based ranges for all provided statuses.

    Algorithm:
    1. For each status, convert its TimeWindowPartitionsSubset to ranges
    2. Combine all ranges into single list
    3. Sort by start time

    Args:
        subsets: Mapping from status enum to TimeWindowPartitionsSubset

    Returns:
        List of TimePartitionRange objects sorted by start time
    """
    all_ranges = []

    # Convert each status subset to ranges
    for status, subset in subsets.items():
        if subset:
            all_ranges.extend(convert_time_window_subset_to_generic_ranges(subset, status))

    # Sort by start time
    all_ranges.sort(key=lambda r: r.start_time)

    return all_ranges


def extract_partition_keys_by_status(
    subsets: Mapping[StatusT, PartitionsSubset | None],
    partitions_def: PartitionsDefinition,
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Mapping[StatusT, Sequence[str]]:
    """Extract and sort partition keys for each status.

    Args:
        subsets: Mapping from status enum to PartitionsSubset (can contain None values)
        partitions_def: Partition definition for creating empty subsets
        dynamic_partitions_store: Store for loading dynamic partitions

    Returns:
        Mapping from status to sorted list of partition keys
    """
    with partition_loading_context(dynamic_partitions_store=dynamic_partitions_store):
        result = {}

        for status, subset in subsets.items():
            if subset is None:
                result[status] = []
            else:
                result[status] = sorted(subset.get_partition_keys())

        return result


@record
class MultiPartitionRange(Generic[StatusT]):
    """Generic multi-partition range (status-agnostic).

    Represents a range in the primary dimension with associated secondary
    dimension statuses (which can be recursive).
    """

    primary_dim_start_key: str
    primary_dim_end_key: str
    primary_dim_start_time: float | None
    primary_dim_end_time: float | None
    secondary_dim: Any  # Recursive structure - can be any partition status type


def build_multi_partition_ranges_generic(
    subsets: Mapping[StatusT, PartitionsSubset | None],
    partitions_def: MultiPartitionsDefinition,
    dynamic_partitions_store: DynamicPartitionsStore,
    build_secondary_dim: Callable[[Mapping[StatusT, PartitionsSubset]], Any],
) -> Sequence[MultiPartitionRange]:
    """Build 2D run-length encoded partition ranges.

    Algorithm:
    1. For each status, build mapping: primary_key → secondary_subset
    2. Run-length encoding:
       - Iterate through primary dimension keys in order
       - Group consecutive primary keys where ALL secondary patterns match
       - For each group, call build_secondary_dim on secondary dimension subsets
    3. Calculate time bounds if primary dimension is time-based

    Args:
        subsets: Mapping from status enum to PartitionsSubset
        partitions_def: Multi-dimensional partition definition
        dynamic_partitions_store: Store for loading dynamic partitions
        build_secondary_dim: Function to recursively build secondary dimension statuses

    Returns:
        List of MultiPartitionRange objects
    """
    # Get dimension definitions
    primary_dim = partitions_def.primary_dimension
    secondary_dim = partitions_def.secondary_dimension

    # Build mappings: primary_key → {status → secondary_keys}
    with partition_loading_context(dynamic_partitions_store=dynamic_partitions_store):
        primary_keys = list(primary_dim.partitions_def.get_partition_keys())

        # For each status, build mapping: primary_key → set of secondary keys
        dim2_keys_by_dim1_by_status: dict[StatusT, dict[str, set[str]]] = {}

        for status, subset in subsets.items():
            if subset:
                dim2_keys_by_dim1 = defaultdict(set)
                for partition_key in subset.get_partition_keys():
                    multipartition_key = partitions_def.get_partition_key_from_str(partition_key)
                    primary_key_val = multipartition_key.keys_by_dimension[primary_dim.name]
                    secondary_key_val = multipartition_key.keys_by_dimension[secondary_dim.name]
                    dim2_keys_by_dim1[primary_key_val].add(secondary_key_val)
                dim2_keys_by_dim1_by_status[status] = dim2_keys_by_dim1

        # Convert to subsets for each primary key
        primary_to_status_subsets: dict[str, dict[StatusT, PartitionsSubset]] = {}

        for primary_key in primary_keys:
            status_subsets: dict[StatusT, PartitionsSubset] = {}

            # Create secondary subsets for each status
            for status, dim2_keys_by_dim1 in dim2_keys_by_dim1_by_status.items():
                if primary_key in dim2_keys_by_dim1:
                    secondary_keys = dim2_keys_by_dim1[primary_key]
                    status_subsets[status] = (
                        secondary_dim.partitions_def.empty_subset().with_partition_keys(
                            secondary_keys
                        )
                    )

            primary_to_status_subsets[primary_key] = status_subsets

        # Run-length encoding: group consecutive primary keys with identical secondary patterns
        ranges = []
        i = 0
        while i < len(primary_keys):
            start_key = primary_keys[i]
            start_subsets = primary_to_status_subsets[start_key]

            # Skip empty primary keys (no partitions in any status)
            if not start_subsets:
                i += 1
                continue

            # Find end of range with identical secondary patterns
            j = i + 1
            while j < len(primary_keys):
                if primary_to_status_subsets[primary_keys[j]] != start_subsets:
                    break
                j += 1

            end_key = primary_keys[j - 1]

            # Build secondary dimension status recursively
            secondary_statuses = build_secondary_dim(start_subsets)

            # Get time range for primary dimension if it's time-based
            start_time = None
            end_time = None
            if isinstance(primary_dim.partitions_def, TimeWindowPartitionsDefinition):
                start_time_window = primary_dim.partitions_def.time_window_for_partition_key(
                    start_key
                )
                end_time_window = primary_dim.partitions_def.time_window_for_partition_key(end_key)
                start_time = start_time_window.start.timestamp()
                end_time = end_time_window.end.timestamp()

            ranges.append(
                MultiPartitionRange(
                    primary_dim_start_key=start_key,
                    primary_dim_end_key=end_key,
                    primary_dim_start_time=start_time,
                    primary_dim_end_time=end_time,
                    secondary_dim=secondary_statuses,
                )
            )

            i = j

        return ranges



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

LOGGER_QUERY = """
query LoggerQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      modes {
        name
        loggers {
          name
          description
          configField {
            configType {
              key
              ... on CompositeConfigType {
                fields {
                  name
                  configType {
                    key
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}
"""


# Warning: If _compute_fields_hash changes, verify that the result.data has the same shape/keys/values
# as the existing snapshot and then run update snapshot
def test_mode_fetch_loggers(graphql_context: WorkspaceRequestContext, snapshot):
    selector = infer_job_selector(graphql_context, "loggers_job")
    result = execute_dagster_graphql(graphql_context, LOGGER_QUERY, {"selector": selector})

    assert not result.errors
    assert result.data
    assert result.data["pipelineOrError"]
    assert result.data["pipelineOrError"]["modes"]
    for mode_data in result.data["pipelineOrError"]["modes"]:
        assert mode_data["loggers"]

    snapshot.assert_match(result.data)



import graphene

from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import GraphenePythonError
from dagster_graphql.schema.util import non_null_list


class GrapheneRunsFeedView(graphene.Enum):
    """Configure how runs and backfills are represented in the feed.

    ROOTS: Return root-level runs and backfills
    RUNS: Return runs only, including runs within backfills
    BACKFILLS: Return backfills only
    """

    ROOTS = "ROOTS"
    RUNS = "RUNS"
    BACKFILLS = "BACKFILLS"

    class Meta:
        name = "RunsFeedView"


class GrapheneRunsFeedEntry(graphene.Interface):
    id = graphene.NonNull(graphene.ID)
    runStatus = graphene.Field("dagster_graphql.schema.pipelines.pipeline.GrapheneRunStatus")
    creationTime = graphene.NonNull(graphene.Float)
    startTime = graphene.Float()
    endTime = graphene.Float()
    tags = non_null_list("dagster_graphql.schema.tags.GraphenePipelineTag")
    jobName = graphene.String()
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKey))
    assetCheckSelection = graphene.List(
        graphene.NonNull("dagster_graphql.schema.entity_key.GrapheneAssetCheckHandle")
    )

    class Meta:
        name = "RunsFeedEntry"


class GrapheneRunsFeedConnection(graphene.ObjectType):
    class Meta:
        name = "RunsFeedConnection"

    results = non_null_list(GrapheneRunsFeedEntry)
    cursor = graphene.NonNull(graphene.String)
    hasMore = graphene.NonNull(graphene.Boolean)


class GrapheneRunsFeedConnectionOrError(graphene.Union):
    class Meta:
        types = (GrapheneRunsFeedConnection, GraphenePythonError)
        name = "RunsFeedConnectionOrError"


class GrapheneRunsFeedCount(graphene.ObjectType):
    class Meta:
        name = "RunsFeedCount"

    count = graphene.NonNull(graphene.Int)


class GrapheneRunsFeedCountOrError(graphene.Union):
    class Meta:
        types = (GrapheneRunsFeedCount, GraphenePythonError)
        name = "RunsFeedCountOrError"


types = [GrapheneRunsFeedConnectionOrError, GrapheneRunsFeedCountOrError]



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository_selector


def all_solids_query():
    return """
    query AllSolidsQuery($repositorySelector: RepositorySelector!) {
        repositoryOrError(repositorySelector: $repositorySelector) {
           ... on Repository {
                usedSolids {
                    __typename
                    definition { name }
                    invocations { pipeline { name } solidHandle { handleID } }
                }
            }
        }
    }
    """


def get_solid_query_exists():
    return """
    query SolidsQuery($repositorySelector: RepositorySelector!, $solidName: String!) {
        repositoryOrError(repositorySelector: $repositorySelector) {
            ... on Repository {
                usedSolid(name: $solidName) {
                    definition {
                        name
                        assetNodes {
                            assetKey {
                                path
                            }
                        }
                    }
                }
            }
            ... on PythonError {
                message
                stack
            }
        }
    }
    """


def test_query_all_solids(graphql_context: WorkspaceRequestContext, snapshot):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context, all_solids_query(), variables={"repositorySelector": selector}
    )
    snapshot.assert_match(result.data)


def test_query_get_solid_exists(graphql_context: WorkspaceRequestContext):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context,
        get_solid_query_exists(),
        variables={"repositorySelector": selector, "solidName": "sum_op"},
    )

    assert not result.errors
    print(result.data["repositoryOrError"])  # noqa: T201
    assert result.data["repositoryOrError"]["usedSolid"]["definition"]["name"] == "sum_op"


def test_asset_graph_node_asset_defs(graphql_context: WorkspaceRequestContext):
    # Verify that the node for an asset graph is correctly associated with its assets

    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context,
        get_solid_query_exists(),
        variables={"repositorySelector": selector, "solidName": "hanging_graph"},
    )

    assert not result.errors
    print(result.data["repositoryOrError"])  # noqa: T201
    assert result.data["repositoryOrError"]["usedSolid"]["definition"]["name"] == "hanging_graph"

    nodes = result.data["repositoryOrError"]["usedSolid"]["definition"]["assetNodes"]

    assert len(nodes) == 1
    assert nodes[0]["assetKey"]["path"] == ["hanging_graph"]
